import pygame
import typing
from mili import error
from mili import data as _data
from mili import typing as _typing
from mili import _coreutils
from mili import _richtext

if typing.TYPE_CHECKING:
    from mili import MILI as _MILI

__all__ = ()


class _PushStylesCtx:
    def __init__(self, mili):
        self._mili = mili

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._mili.pop_styles()


class _globalctx:
    _sample_surf = pygame.Surface((10, 10), pygame.SRCALPHA)
    _font_cache = {}
    _component_types: dict[str, str | _typing.ComponentProtocol] = dict.fromkeys(
        ["rect", "circle", "line", "polygon", "line", "text", "image", "image_layer"],
        "builtin",
    )
    _mili_stack = []
    _mili: "_MILI|None" = None
    _update_ids: dict[str, list[typing.Callable]] = {}

    @classmethod
    def _register_update_id(cls, uid, method):
        if uid is None:
            return
        if uid not in cls._update_ids:
            cls._update_ids[uid] = []
        cls._update_ids[uid].append(method)


class _ctx:
    def __init__(self, mili: "_MILI"):
        self._mili = mili
        self._parent: dict[str, typing.Any] = {
            "rect": pygame.Rect(0, 0, 1, 1),
            "abs_rect": pygame.Rect(0, 0, 1, 1),
            "style": {},
            "id": 0,
            "cd": {
                "children": [],
                "children_grid": [],
                "children_fillx": [],
                "children_filly": [],
            },
            "components": [],
            "parent": None,
            "top": False,
            "z": 0,
            "hovered": False,
            "is_parent": True,
        }
        self._coreutils = _coreutils
        self._stack: dict[str, typing.Any] = self._parent
        self._parents_stack = [self._stack]
        self._id = 1
        self._memory: dict[int, dict[str, typing.Any]] = {}
        self._old_data: dict[int, dict[str, typing.Any]] = {}
        self._interaction_cache: dict[int, _data.Interaction] = {}
        self._element: dict[str, typing.Any] = self._parent
        self._canva: pygame.Surface | None = None
        self._canva_rect: pygame.Rect | None = None
        self._abs_hovered: list[dict[str, typing.Any]] = []
        self._started = False
        self._started_pressing_element = None
        self._started_pressing_button = -1
        self._z = 0
        self._offset = pygame.Vector2()
        self._mouse_pos = pygame.Vector2()
        self._mouse_rel = pygame.Vector2()
        self._global_mouse = False
        self._gmouse_pressed = [False] * 5
        self._gmouse_just_pressed = [False] * 5
        self._gmouse_just_released = [False] * 5
        self._get_just_pressed_func: typing.Callable = pygame.mouse.get_just_pressed
        self._get_just_released_func: typing.Callable = pygame.mouse.get_just_released
        self._image_layer_caches: list[_data.ImageLayerCache] = []
        self._cleared = 0
        self._inside_cache = -1
        self._static_cache = False
        self._last_checkpoint = 0
        self._styles = {}
        self._styles_stack = []
        self._prefabs = {}
        self._dummy_interaction = _data.Interaction(
            mili, False, -1, -1, -1, False, False, 0, {}
        )
        self._default_styles = {
            "element": {},
            "rect": {},
            "circle": {},
            "polygon": {},
            "line": {},
            "text": {},
            "image": {},
            "image_layer": {},
        }

    def _get_element(
        self, rect: pygame.typing.RectLike | None, arg_style, did_begin=False
    ) -> tuple[dict[str, typing.Any], "_data.Interaction"]:
        if len(self._styles_stack) <= 0:
            if arg_style is None:
                style = {}
            else:
                style = arg_style.copy()
        else:
            if arg_style is None:
                style = self._styles.get("element", {})
            else:
                style = self._styles.get("element", {}).copy()
                style.update(arg_style)
        old_el = self._memory.get(self._id, None)
        if self._inside_cache != -1 and old_el:
            if self._static_cache:
                self._z += 1
                self._id += 1
                return old_el, self._dummy_interaction
            else:
                ret_interaction = self._get_interaction(old_el, old_el)
                old_el.update(top=False, hovered=False, style=style, components=[])
                self._element = old_el
                self._z += 1
                self._id += 1
                return old_el, ret_interaction
        if rect is None:
            rect = pygame.Rect()
            absrect = pygame.Rect()
        else:
            rect = pygame.Rect(rect)
            absrect = rect.copy()
        parent = self._parent
        if "parent_id" in style:
            if style["parent_id"] in self._memory:
                parent = self._memory[style["parent_id"]]
            elif style["parent_id"] == 0:
                parent = self._stack
        element: dict[str, typing.Any] = {
            "rect": rect,
            "abs_rect": absrect,
            "style": style,
            "id": self._id,
            "components": [],
            "parent": parent,
            "top": False,
            "z": 1,
            "hovered": False,
            "is_parent": did_begin,
            "fillx": False,
            "filly": False,
            "cd": {
                "children": [],
                "children_grid": [],
                "children_fillx": [],
                "children_filly": [],
            }
            if did_begin
            else {},
        }
        cache = None
        if did_begin:
            cache = self._style_val(style, "element", "cache", None)
            if cache is not None:
                element["cache"] = cache
                if not cache._rebuild:
                    element["cd"] = cache._cache
                    element["grid"] = cache._grid
                    self._inside_cache = self._id
                    self._static_cache = cache._static
        cache_size = self._style_val(style, "element", "cache_rect_size", False)
        oldr = None
        if cache_size and old_el:
            oldr = old_el["rect"]
            rect.w = absrect.w = oldr.w
            rect.h = absrect.h = oldr.h
        constraint = self._style_val(style, "element", "size_clamp", None)
        if constraint:
            element["constraint"] = constraint
            rect.w = absrect.w = self._constrain(rect.w, constraint, 0)
            rect.h = absrect.h = self._constrain(rect.h, constraint, 1)
        interaction = self._get_interaction(element, old_el)
        if old_el:
            oldr = old_el["rect"]
            old_el.update(element)
        else:
            old_el = element
        self._memory[self._id] = old_el
        element = old_el
        if parent:
            z = self._z
            cd = parent["cd"]
            if "z" in style:
                z = style["z"]
            if parent["z"] > z:
                z = parent["z"] + len(cd["children"]) + 1
            cd["children"].append(element)
            if not self._style_val(style, "element", "ignore_grid", False):
                cd["children_grid"].append(element)
                fillx = self._style_val(style, "element", "fillx", False)
                filly = self._style_val(style, "element", "filly", False)
                if fillx:
                    element["fillx"] = fillx
                    cd["children_fillx"].append(element)
                    if oldr is not None:
                        rect.w = absrect.w = oldr.w
                if filly:
                    element["filly"] = filly
                    cd["children_filly"].append(element)
                    if oldr is not None:
                        rect.h = absrect.h = oldr.h
                if oldr is not None:
                    if fillx and self._style_val(style, "element", "resizey", False):
                        rect.h = absrect.h = oldr.h
                    if filly and self._style_val(style, "element", "resizex", False):
                        rect.w = absrect.w = oldr.w
            element["z"] = z + 1
        self._id += 1
        self._element = element
        self._z += 1
        if cache is not None:
            if cache._size != rect.size or style != cache._style:
                cache._size = rect.size
                cache._style = style
                cache._rebuild = True
                element["cd"] = {
                    "children": [],
                    "children_grid": [],
                    "children_fillx": [],
                    "children_filly": [],
                }
                cache._cache = element["cd"]
                cache._grid = None
                self._inside_cache = -1
                self._static_cache = False
        return element, interaction

    def _style_val(self, style: dict, major, minor, default):
        return style.get(minor, self._default_styles[major].get(minor, default))

    def _organize_grid(
        self,
        element,
        a,
        oa,
        av,
        oav,
        rect,
        style,
        pada,
        padoa,
        space,
        anchor,
        resizeoa,
        resizea,
        constraint,
        padded_oa,
        padded_a,
        children,
        ai,
        oai,
    ):
        rectav = getattr(rect, av)
        rectoav = getattr(rect, oav)
        grid_align = self._style_val(style, "element", "grid_align", "first")
        _coreutils._check_align_grid(grid_align)
        spacea = _coreutils._abs_perc(
            self._style_val(style, "element", f"grid_space{a}", space),
            rectav,
        )
        spaceoa = _coreutils._abs_perc(
            self._style_val(style, "element", f"grid_space{oa}", space),
            rectoav,
        )
        line_elements = []
        lines = []
        line_size_a = 0
        longest_line_size_a = 0
        current_pos_oa = padoa
        longest_line_size_oa = 0
        minresizea = float("-inf")
        minresizeoa = float("-inf")
        maxresizea = float("inf")
        maxresizeoa = float("inf")
        if constraint:
            if "min" in constraint:
                mina = constraint["min"][ai]
                minoa = constraint["min"][oai]
                if mina is not None:
                    minresizea = mina
                if minoa is not None:
                    minresizeoa = minoa
            if "max" in constraint:
                maxa = constraint["max"][ai]
                maxoa = constraint["max"][oai]
                if maxa is not None:
                    maxresizea = maxa
                if maxoa is not None:
                    maxresizeoa = maxoa

        for child in element["cd"][f"children_fill{a}"]:
            ch_fill_a = child.get(f"fill{a}", False)
            if ch_fill_a is True:
                ch_fill_a = "100"
            fill_a_v = _coreutils._abs_perc(ch_fill_a, rectav)
            setattr(child["rect"], av, fill_a_v)
            self._organize_element(child)

        for child in element["cd"][f"children_fill{oa}"]:
            ch_fill_oa = child.get(f"fill{oa}", False)
            if ch_fill_oa is True:
                ch_fill_oa = "100"
            fill_oa_v = _coreutils._abs_perc(ch_fill_oa, rectoav)
            setattr(child["rect"], oav, fill_oa_v)
            self._organize_element(child)

        for child in children:
            ch_rect = child["rect"]
            add = spacea
            if line_size_a <= 0:
                add = 0
            hypot_length_a = line_size_a + add + getattr(ch_rect, av)

            if hypot_length_a <= padded_a:
                line_elements.append(child)
                line_size_a = hypot_length_a
                if line_size_a > longest_line_size_a:
                    longest_line_size_a = line_size_a
                if getattr(ch_rect, oav) > longest_line_size_oa:
                    longest_line_size_oa = getattr(ch_rect, oav)
            else:
                if (
                    resizea
                    and hypot_length_a + pada * 2 <= maxresizea
                    and hypot_length_a + pada * 2 >= minresizea
                    and hypot_length_a > longest_line_size_a
                ):
                    longest_line_size_a = hypot_length_a
                    setattr(rect, av, hypot_length_a + pada * 2)
                    padded_a = hypot_length_a
                    line_elements.append(child)
                    line_size_a = hypot_length_a
                    if getattr(ch_rect, oav) > longest_line_size_oa:
                        longest_line_size_oa = getattr(ch_rect, oav)
                else:
                    lines.append([line_elements, longest_line_size_oa, line_size_a])
                    line_elements = []

                    line_elements.append(child)
                    longest_line_size_oa = getattr(ch_rect, oav)
                    line_size_a = getattr(ch_rect, av)
                    if current_pos_oa != padoa:
                        current_pos_oa += spaceoa
                    current_pos_oa += longest_line_size_oa

                    if (
                        resizeoa
                        and current_pos_oa + padoa * 2 <= maxresizeoa
                        and current_pos_oa + padoa * 2 >= minresizeoa
                    ):
                        setattr(rect, oav, current_pos_oa + padoa * 2)
                        padded_oa = current_pos_oa

        if len(line_elements) > 0:
            lines.append([line_elements, longest_line_size_oa, line_size_a])
            if current_pos_oa != padoa:
                current_pos_oa += spaceoa
            current_pos_oa += longest_line_size_oa
            if (
                resizeoa
                and current_pos_oa + padoa * 2 <= maxresizeoa
                and current_pos_oa + padoa * 2 >= minresizeoa
            ):
                setattr(rect, oav, current_pos_oa + padoa * 2)
                padded_oa = current_pos_oa

        rectav = getattr(rect, av)
        rectoav = getattr(rect, oav)

        lines_size_oa = current_pos_oa
        current_pos_oa = padoa
        spacing = spaceoa
        if anchor == "max_spacing":
            if len(lines) > 1:
                spacing = max(
                    spaceoa,
                    ((padded_oa - sum([line[1] for line in lines])) / (len(lines) - 1)),
                )
            else:
                anchor = "center"
        if anchor == "center":
            current_pos_oa = rectoav / 2 - lines_size_oa / 2
        if anchor == "last":
            lines = list(reversed(lines))
            current_pos_oa = padded_oa - lines_size_oa

        for elements, biggest_size, line_size in lines:
            current_pos_a = pada
            spacinga = spacea
            if grid_align == "max_spacing":
                if len(elements) > 1:
                    spacinga = max(
                        0,
                        (padded_a - sum([getattr(e["rect"], av) for e in elements]))
                        / (len(elements) - 1),
                    )
                else:
                    grid_align = "center"
            if grid_align == "last":
                current_pos_a = padded_a - line_size
            elif grid_align == "center":
                current_pos_a = rectav / 2 - line_size / 2
            elif grid_align == "first_center":
                current_pos_a = rectav / 2 - longest_line_size_a / 2
            elif grid_align == "last_center":
                current_pos_a = padded_a / 2 + longest_line_size_a / 2 - line_size
            for el in elements:
                el_rect = el["rect"]
                setattr(el_rect, oa, current_pos_oa)
                setattr(el_rect, a, current_pos_a)
                current_pos_a += getattr(el_rect, av)
                current_pos_a += spacinga
            current_pos_oa += biggest_size
            current_pos_oa += spacing

        element["grid"] = {
            f"overflow{oa}": max(0, lines_size_oa - padded_oa),
            f"overflow{a}": max(0, longest_line_size_a - padded_a),
            f"pad{a}": pada,
            f"pad{oa}": padoa,
            "spacing": space,
            f"space{a}": spacea,
            f"space{oa}": spaceoa,
        }
        if (cache := element.get("cache", None)) is not None:
            cache._cache = element["cd"]
            cache._rebuild = False
            cache._grid = element["grid"]

    def _organize_element(self, element: dict[str, typing.Any]):
        if not element["is_parent"]:
            return
        if self._inside_cache != -1:
            return
        style = element["style"]
        rect = element["rect"]
        cd = element["cd"]
        children = cd["children_grid"]
        a = self._style_val(style, "element", "axis", "y")
        if a not in ["x", "y"]:
            raise error.MILIValueError(f"Invalid axis {a}")
        oa = "x" if a == "y" else "y"
        av = "w" if a == "x" else "h"
        oav = "w" if av == "h" else "h"
        ai = 0 if a == "x" else 1
        oai = 1 if a == "x" else 0
        pad = self._style_val(style, "element", "pad", 5)
        pada = self._style_val(style, "element", f"pad{a}", pad)
        padoa = self._style_val(style, "element", f"pad{oa}", pad)
        rectav = getattr(rect, av)
        rectoav = getattr(rect, oav)
        pada, padoa = (
            _coreutils._abs_perc(pada, rectav),
            _coreutils._abs_perc(padoa, rectoav),
        )
        space = _coreutils._abs_perc(
            self._style_val(style, "element", "spacing", 3), rectav
        )
        anchor = self._style_val(style, "element", "anchor", "first")
        def_align = self._style_val(style, "element", "default_align", "first")
        _coreutils._check_anchor(anchor)
        resizeoa = bool(self._style_val(style, "element", f"resize{oa}", False))
        resizea = bool(self._style_val(style, "element", f"resize{a}", False))
        constraint = element.get("constraint", None)
        _filloa, _filla = (
            self._style_val(style, "element", f"fill{oa}", False),
            self._style_val(style, "element", f"fill{a}", False),
        )
        if resizeoa and bool(_filloa) is not False:
            raise error.MILIIncompatibleStylesError(
                f"Cannot have resize{oa} True and fill{oa} not False"
            )
        if resizea and bool(_filla) is not False:
            raise error.MILIIncompatibleStylesError(
                f"Cannot have resize{a} True and fill{a} not False"
            )

        padded_oa = rectoav - padoa * 2
        padded_a = rectav - pada * 2

        layout = self._style_val(style, "element", "layout", "stack")
        if layout == "table":
            # organize table
            ...
        elif layout == "grid" or self._style_val(style, "element", "grid", False):
            self._organize_grid(
                element,
                a,
                oa,
                av,
                oav,
                rect,
                style,
                pada,
                padoa,
                space,
                anchor,
                resizeoa,
                resizea,
                constraint,
                padded_oa,
                padded_a,
                children,
                ai,
                oai,
            )
            return

        elements_with_filloa = cd[f"children_fill{oa}"]
        elements_with_filla = cd[f"children_fill{a}"]
        biggest_oa = 0
        fixed_elements_a = 0

        for el in children:
            el_rect = el["rect"]
            el_filloa = el.get(f"fill{oa}", False)
            el_filla = el.get(f"fill{a}", False)
            el_constraint = el.get("constraint", None)
            if (
                bool(el_filloa) is False
                or (
                    el_constraint is not None
                    and "min" in el_constraint
                    and el_constraint["min"][oai] is not None
                )
            ) and getattr(el_rect, oav) > biggest_oa:
                biggest_oa = getattr(el_rect, oav)
            fixed_elements_a += space
            if bool(el_filla) is False:
                fixed_elements_a += getattr(el_rect, av)
        fixed_elements_a -= space
        available_to_filla = max(0, padded_a - fixed_elements_a)

        if resizeoa and biggest_oa > 0:
            rectoav = self._constrain(biggest_oa + padoa * 2, constraint, oai)
            setattr(rect, oav, rectoav)
            padded_oa = rectoav - padoa * 2
        if resizea and fixed_elements_a > 0:
            rectav = self._constrain(fixed_elements_a + pada * 2, constraint, ai)
            setattr(rect, av, rectav)
            padded_a = rectav - pada * 2
            available_to_filla = 0

        for filloa_el in elements_with_filloa:
            el_rect = filloa_el["rect"]
            el_style = filloa_el["style"]
            filloa = filloa_el[f"fill{oa}"]
            if filloa is True:
                filloa = "100"
            el_filloa = self._constrain(
                _coreutils._abs_perc(filloa, padded_oa),
                filloa_el.get("constraint", None),
                oai,
            )
            prev = getattr(el_rect, oav)
            if prev != el_filloa or self._cleared > 0:
                setattr(el_rect, oav, el_filloa)
                self._organize_element(filloa_el)

        filla_totalsize = 0
        for filla_el in elements_with_filla:
            el_rect = filla_el["rect"]
            el_style = filla_el["style"]
            filla = filla_el[f"fill{a}"]
            if filla is True:
                filla = "100"
            el_filla = _coreutils._abs_perc(
                filla,
                available_to_filla,
            )
            filla_el["filla"] = el_filla
            filla_totalsize += el_filla

        total_a = fixed_elements_a
        if filla_totalsize != 0:
            multiplier = 1
            if filla_totalsize > available_to_filla:
                multiplier = available_to_filla / filla_totalsize
            for filla_el in elements_with_filla:
                el_rect = filla_el["rect"]
                val = self._constrain(
                    filla_el["filla"] * multiplier, filla_el.get("constraint", None), ai
                )
                total_a += val + space
                prev = getattr(el_rect, av)
                if val != prev or self._cleared > 0:
                    setattr(el_rect, av, val)
                    self._organize_element(filla_el)

        current_a = pada
        spacing = space
        if anchor == "max_spacing":
            if len(children) > 1:
                spacing = max(
                    0,
                    (
                        (padded_a - sum([getattr(e["rect"], av) for e in children]))
                        / (len(children) - 1)
                    ),
                )
            else:
                anchor = "center"
        if anchor == "last":
            current_a = rectav - pada - total_a
        if anchor == "center":
            current_a = max(pada, rectav / 2 - total_a / 2)
        for el in children:
            el_rect = el["rect"]
            el_style = el["style"]
            el_align = self._style_val(el_style, "element", "align", def_align)
            _coreutils._check_align(el_align)
            setattr(el_rect, a, current_a)
            current_a += getattr(el_rect, av) + spacing
            if el_align == "first":
                setattr(el_rect, oa, padoa)
            if el_align == "last":
                setattr(el_rect, "right" if a == "y" else "bottom", rectoav - padoa)
            if el_align == "center":
                setattr(el_rect, f"center{oa}", rectoav / 2)

        element["grid"] = {
            f"overflow{oa}": max(0, biggest_oa - padded_oa),
            f"overflow{a}": max(0, total_a - padded_a),
            f"pad{a}": pada,
            f"pad{oa}": padoa,
            "spacing": space,
            "spacex": space,
            "spacey": space,
        }
        if (cache := element.get("cache", None)) is not None:
            cache._cache = element["cd"]
            cache._rebuild = False
            cache._grid = element["grid"]

    def _constrain(self, value, constraint, vi):
        if constraint is None:
            return value
        minc = constraint.get("min", (None, None))[vi]
        maxc = constraint.get("max", (None, None))[vi]
        if minc is not None and value < minc:
            value = minc
        elif maxc is not None and value > maxc:
            value = maxc
        return value

    def _start(self, style, winpos):
        if self._canva is None:
            return
        self._parent = {
            "rect": self._canva.get_rect(),
            "abs_rect": self._canva.get_rect(),
            "style": style,
            "id": 0,
            "cd": {
                "children": [],
                "children_grid": [],
                "children_fillx": [],
                "children_filly": [],
            },
            "components": [],
            "parent": None,
            "top": False,
            "z": 0,
            "hovered": False,
            "is_parent": True,
        }
        self._id = 1
        self._element = self._parent
        self._stack = self._parent
        self._parents_stack = [self._parent]
        self._abs_hovered = []
        self._started = True
        self._z = 0
        mouse_pos = (
            pygame.Vector2(pygame.mouse.get_pos(self._global_mouse)) - self._offset
        )
        if self._global_mouse:
            if winpos:
                mouse_pos -= winpos
            else:
                raise error.MILIStatusError(
                    "The window position must be provided when using the global mouse state"
                )
        self._mouse_rel = mouse_pos - self._mouse_pos
        self._mouse_pos = mouse_pos
        if 0 in self._memory:
            self._get_old_el(self._stack, None)
        self._memory[0] = self._stack

        if self._global_mouse:
            global_pressed = pygame.mouse.get_pressed(5, True)
            for i, (old, new) in enumerate(zip(self._gmouse_pressed, global_pressed)):
                if old == new:
                    self._gmouse_just_pressed[i] = False
                    self._gmouse_just_released[i] = False
                elif old and not new:
                    self._gmouse_just_released[i] = True
                    self._gmouse_just_pressed[i] = False
                elif new and not old:
                    self._gmouse_just_pressed[i] = True
                    self._gmouse_just_released[i] = False
            self._gmouse_pressed = global_pressed

        if (btn := _coreutils._get_first_button(self._get_just_pressed_func())) > -1:
            self._started_pressing_button = btn
        if (
            _coreutils._get_first_button(self._get_just_released_func())
            == self._started_pressing_button
        ):
            self._started_pressing_button = -1
            self._started_pressing_element = None
        for layer_cache in self._image_layer_caches:
            layer_cache._caches_activity = {}
            layer_cache._rendered = False

    def _get_just_pressed(self):
        return self._gmouse_just_pressed

    def _get_just_released(self):
        return self._gmouse_just_released

    def _check_interaction(self, element: dict[str, typing.Any]):
        blocking = self._style_val(element["style"], "element", "blocking", True)
        if not blocking and blocking is not None:
            return
        parent = element["parent"]
        clipdraw = (
            True
            if parent is None
            else self._style_val(parent["style"], "element", "clip_draw", True)
        )
        hover = element["abs_rect"].collidepoint(self._mouse_pos)
        parent_hover = True
        if parent is not None and clipdraw:
            parent_hover = parent["abs_rect"].collidepoint(self._mouse_pos)
        if hover and parent_hover:
            self._abs_hovered.append(element)

    def _get_old_el(self, element, old_el):
        if old_el is None:
            old_el = self._memory[element["id"]]
        self._old_data[element["id"]] = {
            "rect": old_el["rect"],
            "abs_rect": old_el["abs_rect"],
            "components": old_el["components"],
            "children": old_el["cd"].get("children", ()),
            "parent_id": old_el["parent"]["id"] if old_el["parent"] else -1,
            "grid": old_el["grid"] if "grid" in old_el else None,
        }
        return old_el

    def _get_interaction(
        self, element: dict[str, typing.Any], old_el
    ) -> "_data.Interaction":
        if old_el is None:
            return _coreutils._partial_interaction(self, element, False)
        old_el = self._get_old_el(element, old_el)
        absolute_hover = old_el["abs_rect"].collidepoint(self._mouse_pos)
        if old_el["top"]:
            if self._style_val(element["style"], "element", "blocking", True) is None:
                self._dummy_interaction._raw_data = old_el
                self._dummy_interaction._data = None
                return self._dummy_interaction
            if (
                self._started_pressing_button > -1
                and self._started_pressing_element is None
            ):
                self._started_pressing_element = element["id"]
                return _coreutils._total_interaction(
                    self,
                    element,
                    absolute_hover,
                    absolute_hover,
                    False,
                    old_el,
                )
            if (
                self._started_pressing_button > -1
                and element["id"] != self._started_pressing_element
            ):
                return _coreutils._partial_interaction(self, element, absolute_hover)
            return _coreutils._total_interaction(
                self,
                element,
                absolute_hover,
                absolute_hover,
                False,
                old_el,
            )
        else:
            if (
                self._started_pressing_button > -1
                and element["id"] == self._started_pressing_element
            ):
                return _coreutils._total_interaction(
                    self,
                    element,
                    absolute_hover,
                    False,
                    True,
                    old_el,
                )
            return _coreutils._partial_interaction(self, element, absolute_hover)

    def _add_component(self, type, data, arg_style):
        if self._static_cache:
            return
        if len(self._styles_stack) <= 0:
            if arg_style is None:
                style = {}
            else:
                style = arg_style.copy()
        else:
            if arg_style is None:
                style = self._styles.get(type, {}).copy()
            else:
                style = self._styles.get(type, {}).copy()
                style.update(arg_style)
        if type in ["text", "image"]:
            cache = self._style_val(style, type, "cache", None)
            if cache == "auto":
                style["cache"] = (
                    _data.ImageCache.get_next_cache()
                    if type == "image"
                    else _data.TextCache.get_next_cache()
                )
        element = self._element
        if (eid := style.get("element_id", None)) is not None:
            element = self._memory.get(eid, None)
            if element is None:
                raise error.MILIStatusError(
                    f"Cannot add component to inexistent element with ID {eid}"
                )
        if not self._element:
            raise error.MILIStatusError(
                "Cannot add component to previous element if no element was created"
            )
        if type == "text":
            rich = self._style_val(style, "text", "rich", False)
            cache = None
            if rich:
                cache = self._style_val(style, "text", "cache", None)
                if cache is None:
                    raise error.MILIStatusError(
                        "Rich text requires a valid TextCache object for the cache style"
                    )
                _richtext._comp_init(self, data, style, cache, element["rect"].size)
            self._text_resize(element["rect"], data, style, rich, cache)
        compobj = _globalctx._component_types[type]
        if not isinstance(compobj, str):
            compobj.added(self, data, style, element)
        element["components"].append({"type": type, "data": data, "style": style})

    def _get_font(self, style) -> pygame.Font:
        path = self._style_val(style, "text", "name", None)
        size = int(_coreutils._abs_mod(self._style_val(style, "text", "size", 20)))
        key = f"{path}_{size}"
        if key not in _globalctx._font_cache:
            sysfont = self._style_val(style, "text", "sysfont", False)
            func = pygame.font.SysFont if sysfont else pygame.font.Font
            font = func(path, size)
            _globalctx._font_cache[key] = font
        return _globalctx._font_cache[key]

    def _get_font_from(self, path, size, sysfont):
        size = int(_coreutils._abs_mod(size))
        key = f"{path}_{size}"
        if key not in _globalctx._font_cache:
            func = pygame.font.SysFont if sysfont else pygame.font.Font
            font = func(path, size)
            _globalctx._font_cache[key] = font
        return _globalctx._font_cache[key]

    def _draw_comp_rect(self, data, style, el, rect: pygame.Rect):
        if self._canva is None:
            return
        color = self._style_val(style, "rect", "color", "black")
        if color is None:
            return
        ready_rect = self._style_val(style, "rect", "ready_rect", None)
        pad = self._style_val(style, "rect", "pad", 0)
        padx = self._style_val(style, "rect", "padx", pad)
        pady = self._style_val(style, "rect", "pady", pad)
        padx, pady = (
            int(_coreutils._abs_perc(padx, rect.w)),
            int(_coreutils._abs_perc(pady, rect.h)),
        )
        outline = int(
            _coreutils._abs_perc(
                self._style_val(style, "rect", "outline", 0), min(rect.w, rect.h)
            )
        )
        br_style = self._style_val(style, "rect", "border_radius", 7)
        border_radius = _coreutils._border_radius(br_style, rect.w, rect.h)
        if ready_rect:
            draw_rect = ready_rect
        else:
            aspect_ratio = self._style_val(style, "rect", "aspect_ratio", None)
            if aspect_ratio is None:
                draw_rect = rect.inflate(-padx * 2, -pady * 2)
            else:
                align = self._style_val(style, "rect", "align", "center")
                availx, availy = rect.w - padx * 2, rect.h - pady * 2
                rx, ry = availx, availy
                rx = availy * aspect_ratio
                if rx > availx:
                    ry = availx * (1 / aspect_ratio)
                    rx = availx
                draw_rect = pygame.Rect((0, 0), (rx, ry)).move_to(
                    **_coreutils._align_rect(align, rect, padx, pady)
                )
        dash_size = self._style_val(style, "line", "dash_size", None)
        dash_offset = self._style_val(style, "line", "dash_offset", 0)
        _coreutils._draw_rect(
            self._canva,
            color,
            draw_rect,
            outline,
            border_radius,
            dash_size,
            dash_offset,
        )

    def _draw_comp_circle(self, data, style, el, rect: pygame.Rect):
        if self._canva is None:
            return
        color = self._style_val(style, "circle", "color", "black")
        if color is None:
            return
        pad = self._style_val(style, "circle", "pad", 0)
        padx = self._style_val(style, "circle", "padx", pad)
        pady = self._style_val(style, "circle", "pady", pad)
        padx, pady = (
            int(_coreutils._abs_perc(padx, rect.w)),
            int(_coreutils._abs_perc(pady, rect.h)),
        )
        outline = int(
            _coreutils._abs_perc(
                self._style_val(style, "circle", "outline", 0), min(rect.w, rect.h)
            )
        )
        antialias = self._style_val(style, "circle", "antialias", False)
        circle_rect = rect.inflate(-padx * 2, -pady * 2)
        aspect_ratio = self._style_val(style, "circle", "aspect_ratio", None)
        corners = self._style_val(style, "circle", "corners", None)
        if aspect_ratio is not None:
            align = self._style_val(style, "circle", "align", "center")
            rx, ry = circle_rect.w, circle_rect.h
            rx = circle_rect.h * aspect_ratio
            if rx > circle_rect.w:
                ry = circle_rect.w * (1 / aspect_ratio)
                rx = circle_rect.w
            circle_rect = pygame.Rect((0, 0), (rx, ry)).move_to(
                **_coreutils._align_rect(align, rect, padx, pady)
            )
        dash_size = self._style_val(style, "line", "dash_size", None)
        dash_anchor = self._style_val(style, "line", "dash_anchor", 25)
        _coreutils._draw_circle(
            self._canva,
            circle_rect,
            antialias,
            color,
            outline,
            corners,
            dash_size,
            dash_anchor,
        )

    def _draw_comp_polygon(
        self, data: list[tuple[int, int]], style, el, rect: pygame.Rect
    ):
        if self._canva is None:
            return
        color = self._style_val(style, "polygon", "color", "black")
        if color is None:
            return
        points = []
        pad = self._style_val(style, "line", "pad", 0)
        padx = self._style_val(style, "line", "padx", pad)
        pady = self._style_val(style, "line", "pady", pad)
        padx, pady = (
            int(_coreutils._abs_perc(padx, rect.w)),
            int(_coreutils._abs_perc(pady, rect.h)),
        )
        for raw_p in data:
            rx, ry = raw_p
            rx, ry = _coreutils._abs_perc(rx, rect.w), _coreutils._abs_perc(ry, rect.h)
            if padx * 2 < rect.w:
                rx = pygame.math.clamp(rx, -rect.w // 2 + padx, rect.w // 2 - padx)
            if pady * 2 < rect.h:
                ry = pygame.math.clamp(ry, -rect.h // 2 + pady, rect.h // 2 - pady)
            points.append((rect.centerx + rx, rect.centery + ry))
        outline = int(
            _coreutils._abs_perc(
                self._style_val(style, "polygon", "outline", 0), min(rect.w, rect.h)
            )
        )
        pygame.draw.polygon(self._canva, color, points, outline)

    def _draw_comp_line(
        self, data: list[tuple[int, int]], style, el, rect: pygame.Rect
    ):
        if self._canva is None:
            return
        color = self._style_val(style, "line", "color", "black")
        if color is None:
            return
        points = []
        pad = self._style_val(style, "line", "pad", 0)
        padx = self._style_val(style, "line", "padx", pad)
        pady = self._style_val(style, "line", "pady", pad)
        padx, pady = (
            int(_coreutils._abs_perc(padx, rect.w)),
            int(_coreutils._abs_perc(pady, rect.h)),
        )
        for raw_p in data:
            rx, ry = raw_p
            rx, ry = _coreutils._abs_perc(rx, rect.w), _coreutils._abs_perc(ry, rect.h)
            if padx * 2 < rect.w:
                rx = pygame.math.clamp(rx, -rect.w // 2 + padx, rect.w // 2 - padx)
            if pady * 2 < rect.h:
                ry = pygame.math.clamp(ry, -rect.h // 2 + pady, rect.h // 2 - pady)
            points.append((rect.centerx + int(rx), rect.centery + int(ry)))
        if len(points) != 2:
            raise error.MILIValueError("Wrong number of points")
        size = _coreutils._abs_perc(
            self._style_val(style, "line", "size", 1), min(rect.w, rect.h)
        )
        antialias = self._style_val(style, "line", "antialias", False)
        dash_size = self._style_val(style, "line", "dash_size", None)
        dash_offset = self._style_val(style, "line", "dash_offset", 0)
        width = max(1, int(size))
        _coreutils._draw_line(
            self._canva, antialias, width, color, points, dash_size, dash_offset
        )

    def _draw_comp_text(self, data: str, style, el, rect: pygame.Rect):
        if self._canva is None:
            return
        blit_flags = self._style_val(style, "text", "blit_flags", 0)
        align = self._style_val(style, "text", "align", "center")
        padx = self._style_val(style, "text", "padx", 5)
        pady = self._style_val(style, "text", "pady", 3)
        pad = self._style_val(style, "text", "pad", None)
        if pad is not None:
            padx = pady = pad
        padx, pady = (
            _coreutils._abs_perc(padx, rect.w),
            _coreutils._abs_perc(pady, rect.h),
        )
        rich = self._style_val(style, "text", "rich", False)
        if rich:
            cache = self._style_val(style, "text", "cache", False)
            actions = self._style_val(style, "text", "rich_actions", {})
            _richtext._render(self, cache, rect, blit_flags, align, padx, pady, actions)
            return
        font = self._get_font(style)
        cache = style.get("cache", None)
        fontalign = _coreutils._FONT_ALIGNS[
            self._style_val(style, "text", "font_align", pygame.FONT_CENTER)
        ]
        fontdir = _coreutils._FONT_DIRS[
            self._style_val(style, "text", "font_direction", pygame.DIRECTION_LTR)
        ]
        bold = self._style_val(style, "text", "bold", False)
        italic = self._style_val(style, "text", "italic", False)
        underline = self._style_val(style, "text", "underline", False)
        strikethrough = self._style_val(style, "text", "strikethrough", False)
        antialias = self._style_val(style, "text", "antialias", True)
        color = self._style_val(style, "text", "color", "white")
        bg_color = self._style_val(style, "text", "bg_color", None)
        outline_col = self._style_val(style, "text", "outline_color", None)

        growx, growy = (
            self._style_val(style, "text", "growx", False),
            self._style_val(style, "text", "growy", True),
        )

        wraplen = _coreutils._abs_perc(
            self._style_val(style, "text", "wraplen", 0), rect.w - padx * 2
        )
        if cache is None:
            surf = _coreutils._render_text(
                data,
                font,
                fontalign,
                fontdir,
                antialias,
                color,
                bg_color,
                bold,
                italic,
                underline,
                strikethrough,
                wraplen,
                outline_col,
            )
        else:
            if cache._cache is None:
                new_cache = {
                    "font": font,
                    "fontalign": fontalign,
                    "fontdir": fontdir,
                    "bold": bold,
                    "italic": italic,
                    "underline": underline,
                    "strike": strikethrough,
                    "antialias": antialias,
                    "color": color,
                    "bg_color": bg_color,
                    "outline_col": outline_col,
                    "padx": padx,
                    "pady": pady,
                    "wraplen": wraplen,
                    "data": data,
                }
                surf = _coreutils._render_text(
                    data,
                    font,
                    fontalign,
                    fontdir,
                    antialias,
                    color,
                    bg_color,
                    bold,
                    italic,
                    underline,
                    strikethrough,
                    wraplen,
                    outline_col,
                )
                new_cache["output"] = surf
                cache._cache = new_cache
            else:
                _cache = cache._cache
                if (
                    font is not _cache["font"]
                    or data != _cache["data"]
                    or fontalign != _cache["fontalign"]
                    or fontdir != _cache["fontdir"]
                    or bold != _cache["bold"]
                    or italic != _cache["italic"]
                    or underline != _cache["underline"]
                    or strikethrough != _cache["strike"]
                    or antialias != _cache["antialias"]
                    or color != _cache["color"]
                    or bg_color != _cache["bg_color"]
                    or outline_col != _cache["outline_col"]
                    or padx != _cache["padx"]
                    or pady != _cache["pady"]
                    or wraplen != _cache["wraplen"]
                ):
                    new_cache = {
                        "font": font,
                        "data": data,
                        "fontalign": fontalign,
                        "fontdir": fontdir,
                        "bold": bold,
                        "italic": italic,
                        "underline": underline,
                        "strike": strikethrough,
                        "antialias": antialias,
                        "color": color,
                        "bg_color": bg_color,
                        "outline_col": outline_col,
                        "padx": padx,
                        "pady": pady,
                        "wraplen": wraplen,
                    }
                    surf = _coreutils._render_text(
                        data,
                        font,
                        fontalign,
                        fontdir,
                        antialias,
                        color,
                        bg_color,
                        bold,
                        italic,
                        underline,
                        strikethrough,
                        wraplen,
                        outline_col,
                    )
                    new_cache["output"] = surf
                    cache._cache = new_cache
                else:
                    surf = _cache["output"]
        sw, sh = surf.size
        if growy and sh > rect.h + pady * 2:
            rect.h = sh + pady * 2
        if growx and sw > rect.w + padx * 2:
            rect.w = sw + padx * 2

        txtrect = surf.get_rect(**_coreutils._align_rect(align, rect, padx, pady))
        self._canva.blit(surf, txtrect, special_flags=blit_flags)

    def _text_size(self, data: str, style):
        rich = self._style_val(style, "text", "rich", False)
        if rich:
            cache = self._style_val(style, "text", "cache", None)
            if cache is None:
                raise error.MILIStatusError(
                    "Rich text requires a valid TextCache object for the cache style"
                )
            return cache._rich["size"]
        font = self._get_font(style)
        font.align = _coreutils._FONT_ALIGNS[
            self._style_val(style, "text", "font_align", pygame.FONT_CENTER)
        ]
        font.set_direction(
            _coreutils._FONT_DIRS[
                self._style_val(style, "text", "font_direction", pygame.DIRECTION_LTR)
            ]
        )
        if self._style_val(style, "text", "slow_grow", False):
            wraplen = _coreutils._abs_perc(
                self._style_val(style, "text", "wraplen", 0), 0
            )
            return font.render(
                data, True, "white", None, max(0, int(wraplen))
            ).get_size()
        return font.size(data)

    def _text_resize(self, rect: pygame.Rect, data, style, rich, cache):
        growx, growy = (
            self._style_val(style, "text", "growx", False),
            self._style_val(style, "text", "growy", True),
        )
        if not growx and not growy:
            return
        padx = self._style_val(style, "text", "padx", 5)
        pady = self._style_val(style, "text", "pady", 3)
        pad = self._style_val(style, "text", "pad", None)
        if pad is not None:
            padx = pady = pad
        padx, pady = (
            _coreutils._abs_perc(padx, rect.w),
            _coreutils._abs_perc(pady, rect.h),
        )
        if rich:
            sw, sh = cache._rich["size"]
        else:
            slow_grow = self._style_val(style, "text", "slow_grow", False)
            font = self._get_font(style)
            if slow_grow:
                fontalign = _coreutils._FONT_ALIGNS[
                    self._style_val(style, "text", "font_align", pygame.FONT_CENTER)
                ]
                fontdir = _coreutils._FONT_DIRS[
                    self._style_val(
                        style, "text", "font_direction", pygame.DIRECTION_LTR
                    )
                ]
                wraplen = _coreutils._abs_perc(
                    self._style_val(style, "text", "wraplen", 0), rect.w - padx * 2
                )
                cache = style.get("cache", None)
                if cache is not None and cache._cache is not None:
                    _cache = cache._cache
                    if (
                        font == _cache["font"]
                        and data == _cache["data"]
                        and fontalign == _cache["fontalign"]
                        and fontdir == _cache["fontdir"]
                        and padx == _cache["padx"]
                        and pady == _cache["pady"]
                        and wraplen == _cache["wraplen"]
                    ):
                        sw, sh = cache._cache["output"].get_size()
                    else:
                        font.align = fontalign
                        font.set_direction(fontdir)

                        sw, sh = font.render(
                            data, True, "white", None, max(0, int(wraplen))
                        ).get_size()
                else:
                    font.align = fontalign
                    font.set_direction(fontdir)

                    sw, sh = font.render(
                        data, True, "white", None, max(0, int(wraplen))
                    ).get_size()
            else:
                sw, sh = font.size(str(data))
        if growy and sh + pady * 2 > rect.h + pady * 2:
            rect.h = sh + pady * 2
        if growx and sw + padx * 2 > rect.w + padx * 2:
            rect.w = sw + padx * 2

    def _draw_comp_image(self, data: pygame.Surface, style, el, rect: pygame.Rect):
        if data is None or self._canva is None:
            return
        ready = self._style_val(style, "image", "ready", False)
        blit_flags = self._style_val(style, "image", "blit_flags", 0)
        if ready:
            output = data
        else:
            cache = style.get("cache", None)
            layer_cache = style.get("layer_cache", None)
            pad = self._style_val(style, "image", "pad", 0)
            padx = self._style_val(style, "image", "padx", pad)
            pady = self._style_val(style, "image", "pady", pad)
            padx, pady = (
                _coreutils._abs_perc(padx, rect.w),
                _coreutils._abs_perc(pady, rect.h),
            )
            do_fill = self._style_val(style, "image", "fill", False)
            do_stretchx = self._style_val(style, "image", "stretchx", False)
            do_stretchy = self._style_val(style, "image", "stretchy", False)
            fill_color = self._style_val(style, "image", "fill_color", None)
            smoothscale = self._style_val(style, "image", "smoothscale", True)
            border_radius = self._style_val(style, "image", "border_radius", 0)
            border_radius = _coreutils._border_radius(border_radius, rect.w, rect.h)
            alpha = self._style_val(style, "image", "alpha", 255)
            nine_patch = _coreutils._abs_perc(
                self._style_val(style, "image", "ninepatch_size", 0),
                min(rect.w, rect.h),
            )
            nine_patch = pygame.math.clamp(nine_patch, 0, min(rect.w, rect.h) / 2)
            transforms = self._style_val(style, "image", "transforms", None)
            filters = self._style_val(style, "image", "filters", None)
            if nine_patch != 0 and (do_fill or do_stretchx or do_stretchy):
                raise error.MILIIncompatibleStylesError(
                    "9-patch image mode is incompatible with fill, stretchx, and stretchy styles"
                )
            if layer_cache is not None and cache is None:
                raise error.MILIIncompatibleStylesError(
                    "An ImageCache object must be provided to use the layer_cache style"
                )

            if not cache:
                output = _coreutils._get_image(
                    rect,
                    data,
                    padx,
                    pady,
                    do_fill,
                    do_stretchx,
                    do_stretchy,
                    fill_color,
                    border_radius,
                    alpha,
                    smoothscale,
                    nine_patch,
                    transforms,
                    filters,
                )
                if layer_cache:
                    layer_cache._dirty = True
            else:
                if cache._cache is None:
                    new_cache = {
                        "data": data,
                        "padx": padx,
                        "pady": pady,
                        "fill": do_fill,
                        "stretchx": do_stretchx,
                        "stretchy": do_stretchy,
                        "fill_color": fill_color,
                        "border_radius": border_radius,
                        "alpha": alpha,
                        "size": rect.size,
                        "smoothscale": smoothscale,
                        "nine_patch": nine_patch,
                        "pos": None,
                        "transforms": transforms,
                        "filters": filters,
                    }
                    output = _coreutils._get_image(
                        rect,
                        data,
                        padx,
                        pady,
                        do_fill,
                        do_stretchx,
                        do_stretchy,
                        fill_color,
                        border_radius,
                        alpha,
                        smoothscale,
                        nine_patch,
                        transforms,
                        filters,
                    )
                    new_cache["output"] = output
                    cache._cache = new_cache
                    if layer_cache:
                        layer_cache._dirty = True
                else:
                    _cache = cache._cache
                    if (
                        data is not _cache["data"]
                        or smoothscale != _cache["smoothscale"]
                        or rect.size != _cache["size"]
                        or padx != _cache["padx"]
                        or pady != _cache["pady"]
                        or do_fill != _cache["fill"]
                        or do_stretchx != _cache["stretchx"]
                        or do_stretchy != _cache["stretchy"]
                        or fill_color != _cache["fill_color"]
                        or border_radius != _cache["border_radius"]
                        or alpha != _cache["alpha"]
                        or nine_patch != _cache["nine_patch"]
                        or transforms != _cache["transforms"]
                        or filters != _cache["filters"]
                    ):
                        new_cache = {
                            "data": data,
                            "padx": padx,
                            "pady": pady,
                            "fill": do_fill,
                            "stretchx": do_stretchx,
                            "stretchy": do_stretchy,
                            "fill_color": fill_color,
                            "border_radius": border_radius,
                            "alpha": alpha,
                            "size": rect.size,
                            "smoothscale": smoothscale,
                            "nine_patch": nine_patch,
                            "transforms": transforms,
                            "filters": filters,
                            "pos": None,
                        }
                        output = _coreutils._get_image(
                            rect,
                            data,
                            padx,
                            pady,
                            do_fill,
                            do_stretchx,
                            do_stretchy,
                            fill_color,
                            border_radius,
                            alpha,
                            smoothscale,
                            nine_patch,
                            transforms,
                            filters,
                        )
                        new_cache["output"] = output
                        cache._cache = new_cache
                        if layer_cache:
                            layer_cache._dirty = True
                    else:
                        output = _cache["output"]
        image_rect = output.get_rect(center=rect.center)
        if ready or not layer_cache:
            self._canva.blit(output, image_rect, special_flags=blit_flags)
        else:
            prev = cache._cache["pos"]
            new = pygame.Vector2(image_rect.topleft)
            if new != prev:
                layer_cache._dirty = True
            cache._cache["pos"] = new
            if cache not in layer_cache._caches_set:
                layer_cache._caches_set.add(cache)
                layer_cache._caches.append(cache)
            layer_cache._caches_activity[cache] = True

    def _draw_comp_image_layer(self, data: _data.ImageLayerCache, style, el, rect):
        if self._canva is None:
            return
        clip = self._canva.get_clip()
        self._canva.set_clip(self._canva_rect)
        _coreutils._render_layer_cache(data, self._canva)
        self._canva.set_clip(clip)

    def _draw_update_element(
        self, element: dict[str, typing.Any], parent_pos, parent_clip=None
    ):
        if self._canva is None:
            return
        el_style = element["style"]
        offset = self._style_val(el_style, "element", "offset", (0, 0))
        do_clip = self._style_val(el_style, "element", "clip_draw", True)
        absolute_rect = element["rect"].move(
            (parent_pos[0] + offset[0], parent_pos[1] + offset[1])
        )
        element["abs_rect"] = absolute_rect
        if absolute_rect.w <= 0 or absolute_rect.h <= 0:
            return
        if parent_clip is None:
            parent_clip = absolute_rect
        if not absolute_rect.colliderect(parent_clip):
            return
        if do_clip:
            clip = absolute_rect.clip(parent_clip)
        else:
            clip = parent_clip
        self._check_interaction(element)
        self._canva.set_clip(clip)
        render_above = []
        for component in element["components"]:
            comp_type = component["type"]
            if self._style_val(component["style"], comp_type, "draw_above", False):
                render_above.append(component)
                continue
            compobj = _globalctx._component_types[comp_type]
            if isinstance(compobj, str):
                getattr(self, f"_draw_comp_{comp_type}")(
                    component["data"], component["style"], element, absolute_rect
                )
            else:
                compobj.draw(
                    self, component["data"], component["style"], element, absolute_rect
                )
        cd = element["cd"]
        if len(cd.get("children", ())) > 0:
            for child in sorted(cd["children"], key=lambda c: c["z"]):
                self._canva.set_clip(clip)
                self._draw_update_element(child, absolute_rect.topleft, clip)
                self._canva.set_clip(clip)
        for component in render_above:
            comp_type = component["type"]
            compobj = _globalctx._component_types[comp_type]
            if isinstance(compobj, str):
                getattr(self, f"_draw_comp_{comp_type}")(
                    component["data"], component["style"], element, absolute_rect
                )
            else:
                compobj.draw(
                    self, component["data"], component["style"], element, absolute_rect
                )

    def _start_check(self):
        if not self._started:
            raise error.MILIStatusError("MILI.start() was not called")



import pygame
import typing
import threading
import webbrowser
import copy

from mili import _core
from mili import icon as _icon
from mili import _coreutils
from mili import _richtext
from mili import error as _error
from mili import data as _data
from mili import typing as _typing
from mili._utils._interaction import InteractionCursor as _InteractionCursor


__all__ = (
    "MILI",
    "MarkDown",
    "register_custom_component",
    "pack_component",
    "get_font_cache",
    "clear_font_cache",
    "set_number_modifier",
    "smart_number",
)


def set_number_modifier(
    key_char: str, modifier: typing.Callable[[float], float] | None
):
    if modifier is None:
        if key_char in _coreutils._number_mods:
            del _coreutils._number_mods[key_char]
        return
    if len(key_char) > 1:
        raise _error.MILIValueError(
            "The number modifier character must be a string containing only one character"
        )
    _coreutils._number_mods[key_char] = modifier


def smart_number(smart_number: _typing.SmartNumber):
    return _coreutils._abs_mod(smart_number)


def register_custom_component(name: str, component_type: _typing.ComponentProtocol):
    _core._globalctx._component_types[name] = component_type


def pack_component(name: str, data: typing.Any, style: _typing.AnyStyleLike):
    return {"type": name, "data": data, "style": style}


def get_font_cache():
    return _core._globalctx._font_cache.copy()


def clear_font_cache():
    _core._globalctx._font_cache = {}


def register_update_id(
    update_id: str | None, function: typing.Callable[[_data.Interaction], None]
):
    _core._globalctx._register_update_id(update_id, function)


class MarkDown:
    def __init__(
        self, source: str, style: _typing.MarkDownStyleLike | dict | None = None
    ):
        self._source = source
        self._any_hover = False
        self._parse_result: None | list = None
        self._images = {}
        self._heights = {}
        if style is None:
            style = {}
        copy = style.get("code_copy_style", {})
        sbar = style.get("code_scrollbar_style", {})
        self.style: _typing.MarkDownStyleLike = {
            "quote_width": style.get("quote_width", 3),
            "quote_border_radius": style.get("quote_border_radius", 3),
            "indent_width": style.get("indent_width", 30),
            "separator_height": style.get("separator_height", 5),
            "line_break_size": style.get("line_break_size", 1),
            "bullet_diameter": style.get("bullet_diameter", 5),
            "table_outline_size": style.get("table_outline_size", 1),
            "table_border_radius": style.get("table_border_radius", 7),
            "table_alignx": style.get("table_alignx", "left"),
            "table_aligny": style.get("table_aligny", "center"),
            "line_break_color": style.get("line_break_color", (80, 80, 80)),
            "bullet_color": style.get("bullet_color", "white"),
            "table_bg_color": style.get("table_bg_color", (30, 30, 30)),
            "quote_color": style.get("quote_color", (150, 150, 150)),
            "code_bg_color": style.get("code_bg_color", (40, 40, 40)),
            "code_outline_color": style.get("code_outline_color", (50, 50, 50)),
            "code_border_radius": style.get("code_border_radius", 3),
            "code_copy_style": {
                "size": copy.get("size", 26),
                "icon_color": copy.get(
                    "icon_color",
                    {"default": (120,) * 3, "hover": (180,) * 3, "press": (100,) * 3},
                ),
                "icon_pad": "10",
                "icon": copy.get("icon", _icon.lazy("google", "content_copy")),
            },
            "code_scrollbar_style": {
                "height": sbar.get("height", 7),
                "padx": sbar.get("padx", 1),
                "pady": sbar.get("pady", 1),
                "border_radius": sbar.get("border_radius", 3),
                "bar_color": sbar.get("bar_color", (30,) * 3),
                "handle_color": sbar.get(
                    "handle_color",
                    {
                        "default": (50,) * 3,
                        "hover": (60,) * 3,
                        "press": (50,) * 3,
                    },
                ),
            },
            "title1_size": style.get("title1_size", 26),
            "title2_size": style.get("title2_size", 22),
            "title3_size": style.get("title3_size", 19),
            "title4_size": style.get("title4_size", 16),
            "title5_size": style.get("title5_size", 14),
            "title6_size": style.get("title6_size", 12),
            "allow_image_link": style.get("allow_image_link", True),
            "change_cursor": style.get("change_cursor", True),
            "load_images_async": style.get("load_images_async", True),
            "link_handler": style.get("link_handler", webbrowser.open),
            "parse_async": style.get("parse_async", True),
        }
        self.rebuild()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        changed = False
        if self._source != value:
            changed = True
        self._source = value
        if changed:
            self.rebuild()

    def rebuild(self):
        if self.style["parse_async"]:
            thread = threading.Thread(
                target=_richtext._markdown_parse, args=(self._source, self)
            )
            thread.daemon = True
            thread.start()
        else:
            _richtext._markdown_parse(self._source, self)

    def _link_hover(self, dt):
        self._any_hover = True

    def _link_click(self, dt):
        handler = self.style["link_handler"]
        if handler is None:
            return
        handler(str(dt))


class MILI:
    def __init__(
        self, canva: pygame.Surface | None = None, use_global_mouse: bool = False
    ):
        self._ctx = _core._ctx(self)
        if canva is not None:
            self.canva = canva
        self.use_global_mouse = use_global_mouse
        self.last_interaction: _data.Interaction | None = None
        self.current_parent_interaction: _data.Interaction | None = None

    @property
    def id(self):
        return self._ctx._id

    @id.setter
    def id(self, value):
        self.id_checkpoint(value)

    @property
    def stack_id(self) -> int:
        return self._ctx._stack["id"]

    @property
    def current_parent_id(self) -> int:
        return self._ctx._parent["id"] if self._ctx._parent else -1

    @property
    def all_elements_ids(self) -> list[int]:
        return list(self._ctx._memory.keys())

    @property
    def canva(self) -> pygame.Surface | None:
        return self._ctx._canva

    @canva.setter
    def canva(self, v: pygame.Surface):
        self._ctx._canva = v
        self._ctx._canva_rect = v.get_rect()

    @property
    def canva_offset(self) -> pygame.Vector2:
        return self._ctx._offset

    @canva_offset.setter
    def canva_offset(self, v: typing.Sequence[float]):
        self._ctx._offset = pygame.Vector2(v)

    @property
    def use_global_mouse(self) -> bool:
        return self._ctx._global_mouse

    @use_global_mouse.setter
    def use_global_mouse(self, v: bool):
        self._ctx._global_mouse = v
        if v:
            self._ctx._get_just_pressed_func = self._ctx._get_just_pressed
            self._ctx._get_just_released_func = self._ctx._get_just_released
        else:
            self._ctx._get_just_pressed_func = pygame.mouse.get_just_pressed
            self._ctx._get_just_released_func = pygame.mouse.get_just_released

    def default_style(self, type: str, style: _typing.AnyStyleLike):
        if type != "element" and type not in _core._globalctx._component_types:
            raise _error.MILIValueError("Invalid style type")
        else:
            if type not in self._ctx._default_styles:
                self._ctx._default_styles[type] = {}
        self._ctx._default_styles[type].update(style)

    def default_styles(self, **types_styles: _typing.AnyStyleLike):
        for name, value in types_styles.items():
            self.default_style(name, value)

    def reset_default_styles(self, *types: str):
        for tp in types:
            if tp != "element" and tp not in _core._globalctx._component_types:
                raise _error.MILIValueError("Invalid style type")
            self._ctx._default_styles[tp] = {}

    def push_styles(self, **types_styles: _typing.AnyStyleLike):
        self._ctx._styles_stack.append(copy.deepcopy(self._ctx._styles))
        for type_, style in types_styles.items():
            if type_ not in self._ctx._styles:
                self._ctx._styles[type_] = style
            else:
                self._ctx._styles[type_].update(style)
        return _core._PushStylesCtx(self)

    def pop_styles(self):
        try:
            last = self._ctx._styles_stack.pop()
            self._ctx._styles = last
        except IndexError:
            raise _error.MILIStatusError("MILI.pop_styles called too many times")

    def reset_styles(self):
        self._ctx._styles = {}
        self._ctx._styles_stack = []

    def start(
        self,
        style: _typing.ElementStyleLike | None = None,
        is_global=True,
        window_position: pygame.typing.Point | None = None,
    ) -> typing.Literal[True]:
        if self._ctx._canva is None:
            raise _error.MILIStatusError("Canva was not set")
        _core._globalctx._mili_stack.append(self)
        _core._globalctx._mili = self
        if style is None:
            style = {"blocking": False}
        else:
            style = style.copy()
            style["blocking"] = False
        self._ctx._start(style, window_position)
        if is_global:
            _data.ImageCache._preallocated_index = -1
            _data.TextCache._preallocated_index = -1
        return True

    def update_draw(self):
        self._ctx._organize_element(self._ctx._stack)
        self._ctx._started = False
        self._ctx._draw_update_element(self._ctx._stack, (0, 0))
        for layer_cache in self._ctx._image_layer_caches:
            _core._coreutils._render_layer_cache(layer_cache, self._ctx._canva)
        abs_hovered = sorted(self._ctx._abs_hovered, key=lambda e: e["z"], reverse=True)
        if len(abs_hovered) > 0:
            abs_hovered[0]["top"] = True
        if len(_core._globalctx._mili_stack) > 0:
            _core._globalctx._mili_stack.pop()
        if len(_core._globalctx._mili_stack) > 0:
            _core._globalctx._mili = _core._globalctx._mili_stack[-1]
        else:
            _core._globalctx._mili = None
        if self._ctx._cleared > 0:
            self._ctx._cleared -= 1

    def markdown(self, markdown: MarkDown):
        if markdown.style["change_cursor"]:
            if markdown._any_hover:
                _InteractionCursor.set_cursor(pygame.SYSTEM_CURSOR_HAND)
            markdown._any_hover = False
        _richtext._markdown_ui(self, markdown)

    def element(
        self,
        rect: pygame.typing.RectLike | None,
        style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        self._ctx._start_check()
        el, interaction = self._ctx._get_element(rect, style)
        interaction._did_begin = False
        interaction.parent = self.current_parent_interaction
        self.last_interaction = interaction
        if uid := self._ctx._style_val(el["style"], "element", "update_id", None):
            uids = [uid]
            if not isinstance(uid, str):
                uids = uid
            for ui in uids:
                methods = _core._globalctx._update_ids.get(ui, [])
                for meth in methods:
                    meth(interaction)
        return interaction

    def begin(
        self,
        rect: pygame.typing.RectLike | None,
        style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        self._ctx._start_check()
        el, interaction = self._ctx._get_element(rect, style, True)
        interaction._did_begin = True
        interaction.parent = self.current_parent_interaction
        self.last_interaction = self.current_parent_interaction = interaction
        self._ctx._parent = el
        self._ctx._parents_stack.append(el)
        if uid := self._ctx._style_val(el["style"], "element", "update_id", None):
            uids = [uid]
            if not isinstance(uid, str):
                uids = uid
            for ui in uids:
                methods = _core._globalctx._update_ids.get(ui, [])
                for meth in methods:
                    meth(interaction)
        return interaction

    def add_element_style(
        self, style: _typing.ElementStyleLike | None, element_id: int | None = None
    ):
        self._ctx._start_check()
        if style is None:
            return
        if element_id is None:
            if not self._ctx._element:
                raise _error.MILIStatusError(
                    "Cannot add style to previous element if no element was created"
                )
            element = self._ctx._element
        else:
            element = self._ctx._memory.get(element_id, None)
            if element is None:
                raise _error.MILIValueError(
                    f"Cannot add style to inexistent element with ID {element_id}"
                )
        element["style"].update(style)
        if "z" in style:
            element["z"] = style["z"]

    def end(self):
        self._ctx._start_check()
        parent = self._ctx._parent
        if parent and parent["id"] != 0:
            self._ctx._organize_element(parent)
            if parent["id"] == self._ctx._inside_cache:
                self._ctx._inside_cache = -1
                self._ctx._static_cache = False
        else:
            raise _error.MILIStatusError("end() called too many times")
        if len(self._ctx._parents_stack) > 1:
            self._ctx._parents_stack.pop()
            self._ctx._parent = self._ctx._parents_stack[-1]
        else:
            self._ctx._parent = self._ctx._parents_stack[0]
        self._ctx._element = self._ctx._parent
        if self.current_parent_interaction is not None:
            self.current_parent_interaction = self.current_parent_interaction.parent

    def cache_should_create_children(self):
        return self._ctx._inside_cache == -1 or not self._ctx._static_cache

    def rect(self, style: _typing.RectStyleLike | None = None) -> typing.Self:
        self._ctx._add_component("rect", None, style)
        return self

    def rect_element(
        self,
        rect_style: _typing.RectStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.rect(rect_style)
        return data

    def transparet_rect(
        self, style: _typing.TransparentRectStyleLike | None = None
    ) -> typing.Self:
        if style is None:
            style = {}
        self.image(
            _core._globalctx._sample_surf,
            {
                "fill": True,
                "cache": "auto",
                "pad": style.get("pad", 0),
                "padx": style.get("padx", 0),
                "pady": style.get("pady", 0),
                "fill_color": style.get("color", None),
                "alpha": style.get("alpha", 255),
                "border_radius": style.get("border_radius", 0),
            },
        )
        return self

    def transparent_rect_element(
        self,
        transparent_rect_style: _typing.TransparentRectStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.transparet_rect(transparent_rect_style)
        return data

    def circle(self, style: _typing.CircleStyleLike | None = None) -> typing.Self:
        self._ctx._add_component("circle", None, style)
        return self

    def circle_element(
        self,
        circle_style: _typing.CircleStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.circle(circle_style)
        return data

    def polygon(
        self, points: _typing.PointsLike, style: _typing.PolygonStyleLike | None = None
    ) -> typing.Self:
        self._ctx._add_component("polygon", points, style)
        return self

    def polygon_element(
        self,
        points: _typing.PointsLike,
        polygon_style: _typing.PolygonStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.polygon(points, polygon_style)
        return data

    def line(
        self,
        start_end: _typing.PointsLike,
        style: _typing.LineStyleLike | None = None,
    ) -> typing.Self:
        self._ctx._add_component("line", start_end, style)
        return self

    def hline(self, style: _typing.LineStyleLike | None = None) -> typing.Self:
        return self.line([("-50", 0), ("50", 0)], style)

    def vline(self, style: _typing.LineStyleLike | None = None) -> typing.Self:
        return self.line([(0, "-50"), (0, "50")], style)

    def line_element(
        self,
        start_end: _typing.PointsLike,
        line_style: _typing.LineStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.line(start_end, line_style)
        return data

    def vline_element(
        self,
        line_style: _typing.LineStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.vline(line_style)
        return data

    def hline_element(
        self,
        line_style: _typing.LineStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.hline(line_style)
        return data

    def text(
        self, text: str, style: _typing.TextStyleLike | None = None
    ) -> typing.Self:
        self._ctx._add_component("text", str(text), style)
        return self

    def text_element(
        self,
        text: str,
        text_style: _typing.TextStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.text(text, text_style)
        return data

    def text_size(
        self,
        text: str,
        style: _typing.TextStyleLike | None = None,
        padded: bool = False,
    ) -> pygame.Vector2:
        if style is None:
            style = {}
        size = pygame.Vector2(self._ctx._text_size(text, style))
        if not padded:
            return size

        pad: float = self._ctx._style_val(style, "text", "pad", None)  # type: ignore
        if pad is not None:
            padx = pady = pad
        else:
            padx: float = self._ctx._style_val(style, "text", "padx", 5)  # type: ignore
            pady: float = self._ctx._style_val(style, "text", "pady", 3)  # type: ignore
        size.x += padx * 2
        size.y += pady * 2
        return size

    def text_font(self, style: _typing.TextStyleLike | None = None) -> pygame.Font:
        if style is None:
            style = {}
        return self._ctx._get_font(style)

    def image(
        self, surface: pygame.Surface, style: _typing.ImageStyleLike | None = None
    ) -> typing.Self:
        self._ctx._add_component("image", surface, style)
        return self

    def image_element(
        self,
        surface: pygame.Surface,
        image_style: _typing.ImageStyleLike | None = None,
        element_rect: pygame.typing.RectLike | None = None,
        element_style: _typing.ElementStyleLike | None = None,
    ) -> _data.Interaction:
        data = self.element(element_rect, element_style)
        self.image(surface, image_style)
        return data

    def custom_component(
        self, component: str, data: typing.Any, style: dict[str, typing.Any] | None
    ):
        if (
            component not in self._ctx._default_styles
            and component in _core._globalctx._component_types
        ):
            self._ctx._default_styles[component] = {}
        elif component not in _core._globalctx._component_types:
            raise _error.MILIStatusError(
                f"No custom component '{component}' registered"
            )
        self._ctx._add_component(component, data, style)
        return self

    def packed_component(
        self,
        *packed_components: dict[typing.Literal["type", "style", "data"], typing.Any],
    ):
        for comp in packed_components:
            if comp["type"] not in _core._globalctx._component_types:
                _error.MILIStatusError(f"No component '{comp['type']}' exists")
            self._ctx._add_component(comp["type"], comp["data"], comp["style"])
        return self

    def register_prefab(
        self, name: str, prefab_function: typing.Callable[["MILI"], typing.Any]
    ):
        self._ctx._prefabs[name] = prefab_function

    def prefab(self, name: str, *args, **kwargs) -> typing.Any:
        if name not in self._ctx._prefabs:
            raise _error.MILIValueError(
                f"No prefab was registered with name '{name}'. Existing prefabs: {list(self._ctx._prefabs.keys())}"
            )
        return self._ctx._prefabs[name](self, *args, **kwargs)

    def image_layer_renderer(self, layer: _data.ImageLayerCache) -> typing.Self:
        self._ctx._add_component("image_layer", layer, {"draw_above": True})
        return self

    def data_from_id(self, element_id: int) -> _data.ElementData | None:
        self._ctx._start_check()
        if element_id in self._ctx._memory:
            return _core._coreutils._element_data(
                self._ctx, self._ctx._memory[element_id]
            )
        elif element_id == 0:
            return _core._coreutils._element_data(self._ctx, self._ctx._stack)

    def clear_memory(self, keep_ids: list[int] | None = None):
        ctx = self._ctx
        ctx._cleared = 3
        memory_keep = {}
        old_data_keep = {}
        if keep_ids:
            for _id in keep_ids:
                if _id in ctx._memory:
                    memory_keep[_id] = ctx._memory[_id]
                if _id in ctx._old_data:
                    old_data_keep[_id] = ctx._old_data[_id]
        ctx._memory.clear()
        ctx._old_data.clear()
        ctx._interaction_cache.clear()
        if keep_ids:
            ctx._memory.update(memory_keep)
            ctx._old_data.update(old_data_keep)

    def id_checkpoint(self, id_: int):
        if id_ < 1:
            id_ = 1
        if self._ctx._id <= id_:
            self._ctx._id = id_
            self._ctx._last_checkpoint = id_
        else:
            raise _error.MILIStatusError(
                f"The current ID ({self._ctx._id}) already surpassed the checkpoint ({id_}) so it cannot be applied"
            )

    def id_jump(self, amount: int):
        new_id = self._ctx._last_checkpoint + amount
        if new_id >= self._ctx._id:
            self._ctx._id = new_id
            self._ctx._last_checkpoint = new_id
        else:
            raise _error.MILIStatusError(
                f"Jumping from {self._ctx._last_checkpoint} by {amount} IDs does not surpass the current ID ({self._ctx._id}) so it cannot be done"
            )



import pygame
import mili
import os

COL_DEFAULT = (30, 30, 30)
COL_HOVER = (40, 40, 40)
COL_PRESS = (28, 28, 28)
COL_BG = (24, 24, 24)
COL_OUTLINE = (40, 40, 40)
COL_BG_OUTLINE = (40, 40, 40)
ALPHAS = 180, 255, 150
COL_DISABLED = (180,) * 3
PAD = 6


def link_hover(_):
    mili.InteractionCursor.set_status("hover")


def link_click(data):
    import webbrowser

    webbrowser.open(data)


MARKDOWN = """
# Markdown Demo

Almost anything can be customized through the markdown settings or the other settings. Use the `MILI.push_styles` context sorrounding `MILI.markdown` to apply specific styles to rendered markdown, as it is composed of regular mili elements and components.

## Basics

In a more intuitive way compared to HTML, you can do the following: *italic text*, useful for _emphasis_. For stronger focus, you can make text **bold**. Both work with asterisks and underscores.

Other stylings are also supported: you can ~~strike through~~ text, ==Highlighted text== to make it super visible or `inline code`. Inlined code won't parse markdown styles (`*this is not italic*`)

> Markdown also supports quotes.
> > That can be nested.
> > > Like this
> ## Anything can be quoted, like this title
<br>

## Lists and Nesting

Bullet lists are supported with all the prefixes and can be nested:

- Basic items
  - Nested item 1
  - Nested item 2
    - Even deeper
- Another root-level item

You can also nest markdown within them (and pretty much any other markdown feature):

- **bold inside a list**
- A [link](https://example.com) inside a list
- An `inline code` inside a list
- ## A title inside a list

---

## Spoilers, Code

A way to hide text is supported even if not in the markdown standards by using spoiler tags ||like this one||.

---

Use fenced code blocks for full code snippets:

```
def greet(name):
        return f"Hello, {name}!"
```

Ideal when showing multi-line code with proper formatting and indentation. Code can be copied with the copy button and can be scrolled if it overflows.
---

## Images and Links

Images are supported:

![A sunset](https://picsum.photos/400/220)

And can work as links aswell by using the proper syntax:

[![Click me!](https://picsum.photos/401/220)](https://example.com)

Of course, links don't have to be images [like this one](https://example.com).
If you want to put a link without an alternative text, you can do so as long as it starts with http:// or https:// like this one: <https://google.com>

---

## Tables

Tables can be used for organized content and can contain a lot of styles and elements:

| Feature         | Description                       |
| :-------------- | -------------------------------: |
| *Italic*        | Italic text              |
| **Bold**        | Bolded text             |
| `Inline code`   | Code-style formatting            |
| [Link](https://example.com)       | Clickable hyperlinks             |
| ![Img](https://picsum.photos/500/300)     | Embed an image in a table cell   |

---

## Breaks and Headings (all levels)

Line breaks can be inserted using `\\n` or `<br>`,
like this one.<br>
Or this.

***

You can also use `---`, `___`, or `***` to draw horizontal rules.

---

# Level 1
## Level 2
### Level 3
#### Level 4
##### Level 5
###### Level 6

That's pretty much it. This should be able to render pretty much correctly most of github's READMEs.<br>
"""


class MILIUIDemo(mili.UIApp):
    def __init__(self):
        pygame.init()
        print("MILI " + mili.VERSION_STR)
        super().__init__(
            pygame.Window("MILI-ui Demo (incomplete)", (1000, 850), borderless=True),
            {
                "start_style": {"pady": 0, "padx": 0, "spacing": 0},
                "target_framerate": 120,
            },
        )
        self.original_title = self.window.title
        self.window.minimum_size = (500, 300)
        self.adaptive_scaler.height_weight = 0.1
        self.mili.default_styles(
            rect={"border_radius": 0},
            text={
                "growx": True,
                "growy": True,
                "name": "Segoe UI",
                "sysfont": True,
                "cache": "auto",
                "pady": 5,
            },
            image={"cache": "auto", "smoothscale": True},
            circle={"pad": 1},
        )
        self.scroll = mili.Scroll("scroll")
        self.scrollbar = mili.Scrollbar(
            self.scroll,
            {"handle_update_id": "sbar_handle"},
        )
        if not os.path.exists("demo_cache"):
            os.mkdir("demo_cache")
        mili.icon.setup("demo_cache", "white")
        self.sections = {
            "layout": False,
            "text": False,
            "shapes": False,
            "images": False,
            "animations": False,
            "prefab examples": False,
            "markdown": False,
        }
        self.rect_offset = 0
        self.rect_size = 0
        self.rect_space = 0
        self.circle_anchor = 0
        self.checkbox_selectables = [
            mili.Selectable(True if i == 1 else False, f"sel{i}") for i in range(4)
        ]
        self.radio_selectables = [
            mili.Selectable(True if i == 2 else False) for i in range(4)
        ]
        self.menu = mili.ContextMenu(
            self.mili,
            {
                "clamp_size": self.window.size,
                "menu_width": 200,
                "text_style": {"size": "s17"},
            },
        )
        self.menu_area_rect = None
        self.menu_pos = None
        self.markdown = mili.MarkDown(MARKDOWN)
        self.dropdown = mili.DropMenu(
            ["A", "B", "C", "D", "E", "F"],
            "B",
            ui_style={
                "text_style": {"size": "s18", "align": "left"},
                "option_text_style": {"size": "s16", "align": "left"},
            },
        )
        self.dropup = mili.DropMenu(
            ["A", "B", "C", "D", "E", "F"],
            "B",
            {"direction": "up"},
            {
                "menu_style": {"pad": 3, "spacing": 0},
                "option_border_radius": 7,
                "text_style": {"size": "s18", "align": "left"},
                "option_text_style": {"size": "s16", "align": "left"},
            },
        )

        self.CMF_autosave = False
        self.CMF_backups = True
        self.CMF_realtime_sync = False
        self.CMF_smart_edit = True
        self.CMF_multi_cursor = False
        self.CMF_highlight_changes = True
        self.CMF_show_line_numbers = True
        self.CMF_show_minimap = False
        self.CMF_word_wrap = True
        self.CMF_code_folding = False
        self.CMF_theme_dark = True
        self.CMF_theme_light = False
        self.CMF_theme_solarized = False
        self.CMF_theme_custom = False
        self.CMF_auto_compile = True
        self.CMF_show_console = True
        self.CMF_live_reload = False
        self.CMF_devtools = False
        self.CMF_experimental = False
        self.CMF_logs_enabled = True
        self.CMF_verbose_mode = False
        self.CMF_extensions_autoupdate = True
        self.CMF_load_external_plugins = False
        self.CMF_stay_logged_in = True
        self.CMF_cloud_sync = False
        self.CMF_check_updates = True
        self.CMF_telemetry = False
        self.CMF_error_reports = True

    def update(self):
        self.scrollbar.style["short_size"] = self.scale(8)
        self.rect_offset += 0.8
        self.circle_anchor += 0.2
        if self.rect_offset >= self.rect_size + self.rect_space:
            self.rect_offset = 0
        if self.circle_anchor > 10:
            self.circle_anchor = 0
        self.mili.default_styles(
            text={
                "size": "s20",
                "growx": True,
                "growy": True,
                "name": "Segoe UI",
                "sysfont": True,
                "cache": "auto",
                "pady": 3,
            },
        )
        if self.menu_area_rect is not None:
            self.menu.style["clamp_size"] = self.menu_area_rect.size
        self.menu.style["menu_width"] = self.scale(200)
        self.window.title = (
            f"{self.original_title} ({self.clock.get_fps():.0f}/120 FPS)"
        )

    def ui(self):
        needed = self.scrollbar.needed
        sbar_space = (
            self.scrollbar.style["short_size"] + self.scrollbar.style["border_dist"] * 2
        )
        with self.mili.begin(
            None,
            {
                "update_id": "scroll",
                "pady": 0,
                "padx": 0,
                "fillx": True,
                "filly": True,
            },
        ):
            if needed:
                self.ui_scrollbar()
            self.mili.id_checkpoint(1000)
            with self.mili.begin(
                (0, 0, self.window.size[0] - sbar_space - PAD * 2, 0),
                {
                    "resizey": True,
                    "offset": self.scroll.get_offset(),
                    "align": "center",
                    "pad": 0,
                    "default_align": "center",
                },
            ):
                self.mili.text_element(
                    "MILI-ui Demo (incomplete)",
                    {"size": "s35"},
                    None,
                    {"align": "center"},
                )
                self.mili.text_element(
                    "Everything inside this window, including the window movements are done with mili-ui. Check this app's code to see how everything is implemented",
                    {
                        "size": "s16",
                        "color": COL_DISABLED,
                        "slow_grow": True,
                        "wraplen": "100",
                    },
                    None,
                    {"align": "center", "fillx": True},
                )
                self.ui_sections()

    def ui_sections(self):
        self.mili.id_checkpoint(10000)
        for name, active in list(self.sections.items()):
            self.mili.id_jump(1000)
            with self.mili.begin(
                None,
                {
                    "resizey": True,
                    "fillx": True,
                    "update_id": "cursor",
                }
                | mili.CENTER
                | mili.PADLESS
                | mili.SPACELESS
                | mili.X,
            ) as row:
                self.mili.rect(
                    mili.style.color(
                        mili.style.cond_value(
                            row, COL_DEFAULT, COL_HOVER, COL_PRESS, force_hover=active
                        )
                    )
                    | {"border_radius": 7}
                )
                self.ui_arrow(active, row.data.rect.h)
                with self.mili.element(None, {"fillx": "100", "blocking": False}):
                    self.mili.text(name.title(), {"growx": False, "size": "s20"})
                if row.left_clicked and self.win_borders.can_interact():
                    self.sections[name] = not active
                self.ui_arrow(active, row.data.rect.h)
            if active:
                func = getattr(self, f"ui_section_{name.replace(' ', '_')}", None)
                if not func:
                    self.mili.text_element("This section is not implemented yet.")
                else:
                    func()

    def ui_menu_bg(self):
        self.mili.rect({"color": COL_BG})
        self.mili.rect({"outline": 1, "color": COL_OUTLINE, "draw_above": True})

    def ui_section_layout(self):
        self.mili.text_element(
            "The following containers will showcase the flexible ways MILI can organize elements. The gray children could be any element, from single components to infinitely layered containers. The containers can resize to fit their elements and can have customizable spacing and padding.",
            {"wraplen": mili.percentage(90, self.window.size[0]), "slow_grow": True},
        )
        width = self.scale(500)
        height = self.scale(180)
        ew = self.scale(30)
        sqh = self.scale(30)
        with self.mili.begin((0, 0, width, height)):
            self.ui_menu_bg()
            self.mili.text_element(
                "Elements ignoring grid",
                {"align": "center", "color": (150,) * 3, "size": "s17"},
            )
            for pos in [
                (123, 45),
                (298, 102),
                (450, 145),
                (35, 88),
                (210, 52),
                (375, 140),
                (60, 130),
                (180, 55),
                (320, 30),
                (10, 120),
            ]:
                self.mili.rect_element(
                    {"color": COL_HOVER},
                    ((self.scale(pos[0]), self.scale(pos[1])), (ew, ew)),
                    {"ignore_grid": True},
                )
        eh = self.scale(30)
        ew = self.scale(50)
        with self.mili.begin((0, 0, width, 0), {"resizey": True}):
            self.ui_menu_bg()
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, self.scale(200), eh), {"align": "first"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, self.scale(110), eh), {"align": "center"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, self.scale(250), eh), {"align": "last"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, self.scale(150), eh), {"align": "first"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, self.scale(230), eh), {"align": "first"}
            )
        with self.mili.begin(
            (0, 0, width, height), {"axis": "x", "anchor": "max_spacing"}
        ):
            self.ui_menu_bg()
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(85)), {"align": "first"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(145)), {"align": "first"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(55)), {"align": "last"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(95)), {"align": "last"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(100)), {"align": "center"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(160)), {"align": "last"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(40)), {"align": "first"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(85)), {"align": "center"}
            )
            self.mili.rect_element(
                {"color": COL_HOVER}, (0, 0, ew, self.scale(170)), {"align": "center"}
            )
        for anchor in ["center", "first", "last", "max_spacing"]:
            with self.mili.begin(
                (0, 0, width, 0), {"axis": "x", "resizey": True, "anchor": anchor}
            ):
                self.ui_menu_bg()
                for i in range(4):
                    self.mili.rect_element({"color": COL_HOVER}, (0, 0, eh, eh))
        eh = self.scale(80)
        with self.mili.begin((0, 0, width, 0), {"axis": "x", "resizey": True}):
            self.ui_menu_bg()
            with self.mili.element((0, 0, 100, eh)):
                self.mili.rect(
                    {"color": "gray40"},
                )
                self.mili.text("Fixed: 100")
            with self.mili.element(
                (0, 0, 0, eh),
                {"fillx": "100"},
            ):
                self.mili.rect({"color": COL_HOVER})
                self.mili.text("'fillx': '100'")
            with self.mili.element(
                (0, 0, 0, eh),
                {"fillx": "50"},
            ):
                self.mili.rect({"color": COL_HOVER})
                self.mili.text("'fillx': '50'")
            with self.mili.element((0, 0, 50, eh)):
                self.mili.rect(
                    {"color": "gray40"},
                )
                self.mili.text("Fixed: 50")
        with self.mili.element(
            (0, 0, 0, eh), {"fillx": "50", "size_clamp": {"min": (500, None)}}
        ):
            self.mili.rect({"color": "gray40"})
            self.mili.text(
                r"fillx is 50% but is never smaller than 500px (size clamping)",
                {"color": "black"},
            )
        self.mili.text_element("Next use the grid layout with different styling")
        with self.mili.begin(
            (0, 0, width, 0), {"resizey": True, "axis": "x", "layout": "grid"}
        ):
            self.ui_menu_bg()
            for i in range(50):
                self.mili.element((0, 0, sqh, sqh))
                self.mili.rect({"color": COL_HOVER})
                self.mili.text(f"{i}", {"size": "s15"})
        with self.mili.begin((0, 0, width, height), {"layout": "grid"}):
            self.ui_menu_bg()
            for i in range(52):
                self.mili.element((0, 0, sqh, sqh))
                self.mili.rect({"color": COL_HOVER})
                self.mili.text(f"{i}", {"size": "s15"})
        with self.mili.begin(
            (0, 0, width, height),
            {
                "layout": "grid",
                "align": "center",
                "anchor": "max_spacing",
                "grid_align": "center",
            },
        ):
            self.ui_menu_bg()
            for i in range(27):
                self.mili.element((0, 0, sqh, sqh))
                self.mili.rect({"color": COL_HOVER})
                self.mili.text(f"{i}", {"size": "s15"})

    def ui_section_text(self):
        self.mili.text_element("You can easily display text")
        self.mili.text_element("It supports every pygame capability", {"size": "s25"})
        with self.mili.begin(
            None, mili.PADLESS | mili.X | mili.CENTER | mili.SPACELESS | mili.RESIZE
        ):
            self.mili.text_element(
                "Like color or", {"color": "yellow", "bg_color": "red"}
            )
            self.mili.text_element("Bold, ", {"bold": True})
            self.mili.text_element("Italic, ", {"italic": True})
            self.mili.text_element("Striketrough,  ", {"strikethrough": True})
            self.mili.text_element("Underline, ", {"underline": True})
            self.mili.text_element("A different font, ", {"name": "Times New Roman"})
            self.mili.text_element("Not antialiased, ", {"antialias": False})
            self.mili.text_element(
                "Outlined", {"color": "black", "outline_color": "white"}
            )
        self.mili.text_element(
            "And it obviously supports\nmultiline strings\nin the same component\nwith different alignments",
            {
                "font_align": pygame.FONT_RIGHT,
                "slow_grow": True,
                "bg_color": COL_HOVER,
            },
        )
        self.mili.hline_element({"color": COL_HOVER}, (0, 0, 0, 1), {"fillx": True})
        self.mili.text_element(
            f'If you want to apply all kind of <b>different styles</b> to your text in a <i>single element</i> and with <u>properly working</u> text <s>wrapping</s> you can use the custom-made <font size="{self.scale(30)}">rich text mode</font>. It supports tags like <font size="{self.scale(26)}" name="Times New Roman" sysfont>b, i, s, u</font> for <color fg="black" outline="white">text style</color> and can <color fg="green">change font</color> or <color bg="blue">change color.</color> Multiple styles can obviously be <b><i><u><color fg="red" outline="white">combined <font size="{self.scale(20)}">in</color> all kind <color fg="yellow" bg="purple">of </font>different ways</color> by using</b> HTML-like</i> tags</u>',
            {
                "rich": True,
                "wraplen": mili.percentage(self.window.size[0], 85),
                "font_align": "left",
                "align": "left",
            },
        )

        self.mili.hline_element({"color": COL_HOVER}, (0, 0, 0, 1), {"fillx": True})
        self.mili.text_element(
            'The rich text has a few other perks: actions can be performed when certain portions of the text are clicked and certain styles are only applied if certain conditions are met, dinamically. For example, it is possible to add <a href="https://www.google.com">proper links like the current one (it will open google)</a> or easily implement <action name="link_hover" condition="hover"><color fg="gray20" bg="gray20"><color fg="(150, 150, 150)" bg="null" condition="just-released">spoiler tags like the one you just opened</color></color></action>. The same component can also render markdown text styling that is showcased in the below section.',
            {
                "rich": True,
                "rich_actions": {
                    "link_hover": link_hover,
                    "link_click": link_click,
                },
                "wraplen": mili.percentage(self.window.size[0], 85),
            },
        )

    def ui_section_shapes(self):
        self.mili.text_element(
            "MILI can add any shape provided by pygame to any element."
        )
        self.mili.text_element("Lines:")
        self.mili.hline_element({"color": "green"}, (0, 0, 0, 10), {"fillx": True})
        self.mili.hline_element(
            {"color": "yellow", "size": 5}, (0, 0, 0, 10), {"fillx": True}
        )
        self.mili.line_element(
            [("-50", "-50"), ("50", "50")],
            {"size": 2, "color": "red", "pady": 3},
            (0, 0, 0, 40),
            {"fillx": True},
        )
        el = self.mili.line_element(
            [("-50", "50"), ("50", "-50")],
            {"size": 1, "color": "blue", "pady": 3, "antialias": True},
            (0, 0, 0, 40),
            {"fillx": True},
        )
        self.mili.hline_element(
            {
                "size": 3,
                "dash_size": [el.data.rect.w / 10, el.data.rect.w / 20],
                "color": "white",
            },
            (0, 0, 0, 10),
            {"fillx": True},
        )
        self.mili.hline_element(
            {
                "size": 3,
                "dash_size": [el.data.rect.w / 100, el.data.rect.w / 200],
                "color": "pink",
            },
            (0, 0, 0, 10),
            {"fillx": True},
        )
        self.mili.text_element("Rects:")
        width = self.scale(200)
        height = self.scale(120)
        rrect = (0, 0, width, height)
        perimeter = width * 2 + height * 2
        perimeter = perimeter / 20
        self.rect_size = perimeter * (8 / 10)
        self.rect_space = perimeter * (2 / 10)
        with self.mili.begin(
            None,
            {
                "fillx": True,
                "resizey": True,
                "layout": "grid",
                "axis": "x",
                "padx": 10,
                "grid_align": "center",
                "default_align": "center",
            },
        ):
            self.mili.rect_element({"color": "gray30"}, rrect)
            self.mili.rect_element({"color": "white", "outline": 1}, rrect)
            self.mili.rect_element({"color": "green", "outline": "10"}, rrect)
            self.mili.rect_element(
                {"color": "yellow", "outline": 5, "border_radius": "50"}, rrect
            )
            self.mili.rect_element({"color": "pink", "border_radius": "30"}, rrect)
            self.mili.rect_element(
                {"color": "yellow", "outline": 1, "border_radius": 20}, rrect
            )
            self.mili.rect_element(
                {"color": "red", "outline": 2, "border_radius": [0, 20, "10", 50]},
                rrect,
            )
            with self.mili.element(rrect):
                self.mili.rect({"color": "gray20", "outline": 1})
                self.mili.rect({"color": "white", "aspect_ratio": 1, "pady": 10})
            self.mili.rect_element(
                {"color": "white", "outline": 2, "dash_size": ("2", "1")}, rrect
            )
            self.mili.rect_element(
                {
                    "color": "green",
                    "outline": 10,
                    "dash_size": (str(100 / 16), str(100 / 32)),
                },
                rrect,
            )
            self.mili.rect_element(
                {
                    "color": "blue",
                    "outline": 5,
                    "dash_size": (self.rect_size, self.rect_space),
                    "dash_offset": self.rect_offset,
                    "aspect_ratio": 1,
                },
                rrect,
            )
        self.mili.text_element("Circles/Ellipses:")
        width = self.scale(150)
        height = self.scale(150)
        crect = (0, 0, width, height)
        with self.mili.begin(
            None,
            {
                "fillx": True,
                "resizey": True,
                "layout": "grid",
                "axis": "x",
                "padx": 10,
                "grid_align": "center",
                "default_align": "center",
            },
        ):
            self.mili.circle_element({"color": "gray50"}, crect)
            self.mili.circle_element({"color": "gray50", "antialias": True}, crect)
            self.mili.circle_element(
                {"color": "white", "antialias": True, "outline": 1}, crect
            )
            with self.mili.push_styles(circle={"antialias": True}):
                self.mili.circle_element({"color": "yellow", "outline": "10"}, crect)
                self.mili.circle_element(
                    {"color": "green", "corners": [True, False, True, False]}, crect
                )
                self.mili.circle_element(
                    {
                        "color": "green",
                        "corners": [False, True, False, True],
                        "outline": 20,
                    },
                    crect,
                )
                self.mili.circle_element(
                    {"color": "red", "outline": 1, "padx": "1", "pady": "10"}, crect
                )
                self.mili.circle_element(
                    {"color": "red", "padx": "10", "pady": "1"}, crect
                )
                self.mili.circle_element(
                    {"color": "pink", "dash_size": (4, 2), "outline": 2}, crect
                )
                self.mili.circle_element(
                    {"color": "green", "dash_size": (5, 5), "outline": 1}, crect
                )
                self.mili.circle_element(
                    {
                        "color": "red",
                        "dash_size": (5, 5),
                        "dash_anchor": self.circle_anchor,
                        "outline": 1,
                    },
                    crect,
                )
                self.mili.circle_element(
                    {
                        "color": "blue",
                        "dash_size": (5, 5),
                        "dash_anchor": self.circle_anchor,
                        "outline": 1,
                        "pady": "15",
                    },
                    crect,
                )
        self.mili.text_element("Polygons:")
        width = self.scale(200)
        height = self.scale(200)
        prect = (0, 0, width, height)
        points = [
            ("0", "-50"),
            ("14", "-16"),
            ("47", "-16"),
            ("23", "6"),
            ("29", "40"),
            ("0", "20"),
            ("-29", "40"),
            ("-23", "6"),
            ("-47", "-16"),
            ("-14", "-16"),
        ]
        with self.mili.begin(
            None,
            {
                "fillx": True,
                "resizey": True,
                "layout": "grid",
                "axis": "x",
                "padx": 10,
                "grid_align": "center",
                "default_align": "center",
            },
        ):
            self.mili.polygon_element(points, {"color": "green"}, prect)
            self.mili.polygon_element(points, {"color": "yellow", "outline": 2}, prect)

    def ui_section_prefab_examples(self):
        self.mili.push_styles(rect={"border_radius": 7}, circle={"antialias": True})
        self.mili.text_element(
            "All the examples here use objects from the mili.utility submodule, whose members are exported to the mili namespace. The behavior of this window exists because of them, as well as the cursor changing. You can already see the Scroll/Scrollbar utility in action.",
            {"wraplen": mili.percentage(90, self.window.size[0])},
        )
        self.mili.text_element("Selectables:")
        size = self.scale(30)
        with self.mili.begin(
            None, {"fillx": "80", "resizey": True, "axis": "x", "anchor": "max_spacing"}
        ):
            self.ui_menu_bg()
            for i, selectable in enumerate(self.checkbox_selectables):
                with self.mili.begin(
                    None, {"resizex": True, "resizey": True, "axis": "x", "pad": 0}
                ):
                    with self.mili.element(
                        (0, 0, size, size), {"update_id": [f"sel{i}", "cursor"]}
                    ):
                        self.mili.rect(mili.style.outline(COL_OUTLINE))
                        if selectable.selected:
                            self.mili.rect({"color": "blue", "pad": 3})
                    self.mili.text_element(f"Checkbox {i}")
        with self.mili.begin(
            None, {"fillx": "80", "resizey": True, "axis": "x", "anchor": "max_spacing"}
        ):
            self.ui_menu_bg()
            for i, selectable in enumerate(self.radio_selectables):
                with self.mili.begin(
                    None, {"resizex": True, "resizey": True, "axis": "x", "pad": 0}
                ):
                    with selectable.update(
                        self.mili.element((0, 0, size, size), {"update_id": "cursor"}),
                        self.radio_selectables,
                        False,
                    ):
                        self.mili.circle(mili.style.outline(COL_OUTLINE))
                        if selectable.selected:
                            self.mili.circle({"color": "blue", "pad": 4})
                    self.mili.text_element(f"Radio Button {i}")
        self.mili.text_element("Sliders (+Dragger):")
        self.mili.text_element("Drop Menus:")
        self.dropup.ui((0, 0, 0, self.scale(35)), {"fillx": "50"})
        self.dropdown.ui((0, 0, 0, self.scale(35)), {"fillx": "50"})
        self.mili.text_element("Entrylines:")
        self.mili.pop_styles()
        self.mili.text_element("Context Menu (right click in the box below):")
        with self.mili.begin((0, 0, 0, self.scale(600)), {"fillx": True}) as cont:
            self.menu_area_rect = cont.data.absolute_rect
            self.ui_menu_bg()
            if self.menu.opened:
                self.ui_context_menu(cont.data.id)

    def ui_section_markdown(self):
        with self.mili.push_styles(
            text={"align": "left", "font_align": "left"},
            element={"align": "first", "anchor": "first"},
        ):
            self.mili.markdown(self.markdown)

    def ui_arrow(self, active, h):
        self.mili.image_element(
            mili.icon.get_google("arrow_drop_up" if active else "arrow_drop_down"),
            None,
            (0, 0, h, 0),
            {"blocking": False, "filly": True},
        )

    def ui_scrollbar(self):
        with self.mili.begin(self.scrollbar.bar_rect, self.scrollbar.bar_style):
            with self.mili.element(
                self.scrollbar.handle_rect,
                self.scrollbar.handle_style | {"update_id": ["cursor", "sbar_handle"]},
            ) as handle:
                self.mili.rect(
                    {
                        "border_radius": 5,
                        **mili.style.color(
                            mili.style.cond_value(
                                handle, COL_DEFAULT, COL_HOVER, COL_PRESS
                            )
                        ),
                    }
                )

    def event(self, event):
        self.scroll.wheel_event(event)
        if not self.menu_area_rect:
            return
        if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == pygame.BUTTON_RIGHT:
                self.menu.close()
                if self.menu_area_rect.collidepoint(event.pos):
                    self.menu.open()
                    self.menu_pos = (
                        pygame.Vector2(event.pos) - self.menu_area_rect.topleft
                    )
            elif event.button == pygame.BUTTON_LEFT and not self.menu.collidepoint(
                pygame.Vector2(event.pos) - self.menu_area_rect.topleft
            ):
                self.menu.close()

    def ui_context_menu(self, parent):
        if not self.menu_pos:
            return
        with self.menu.begin(self.menu_pos, parent):
            self.menu.icon_label("Project Control Center", "psychology")
            if self.menu.label_submenu("File"):
                with self.menu.begin_submenu():
                    self.menu.label("New")
                    self.menu.label("Open")
                    self.menu.label("Open Recent")
                    self.menu.label("Save")
                    self.menu.label("Save As...")
                    self.menu.label("Revert")
                    self.menu.separator()
                    change, self.CMF_autosave = self.menu.checkbox(
                        "Enable Autosave", self.CMF_autosave
                    )
                    change, self.CMF_backups = self.menu.checkbox(
                        "Create Backup on Save", self.CMF_backups
                    )
                    change, self.CMF_realtime_sync = self.menu.checkbox(
                        "Real-Time Sync", self.CMF_realtime_sync
                    )
                    self.menu.separator()
                    if self.menu.label_submenu("Templates"):
                        with self.menu.begin_submenu():
                            self.menu.label("HTML5")
                            self.menu.label("Vue Project")
                            self.menu.label("React Project")
                            self.menu.label("CLI Python Tool")
                            if self.menu.label_submenu("Backend Frameworks"):
                                with self.menu.begin_submenu():
                                    self.menu.label("Flask API")
                                    self.menu.label("Django Site")
                                    self.menu.label("FastAPI")
                                    if self.menu.label_submenu("Django Variants"):
                                        with self.menu.begin_submenu():
                                            self.menu.label("Admin Only")
                                            self.menu.label("API-Only")
                                            self.menu.label("Fullstack")

            if self.menu.label_submenu("Edit"):
                with self.menu.begin_submenu():
                    self.menu.label("Undo")
                    self.menu.label("Redo")
                    self.menu.label("Cut")
                    self.menu.label("Copy")
                    self.menu.label("Paste")
                    self.menu.label("Duplicate Line")
                    change, self.CMF_smart_edit = self.menu.checkbox(
                        "Enable Smart Edit", self.CMF_smart_edit
                    )
                    change, self.CMF_multi_cursor = self.menu.checkbox(
                        "Multi-Cursor Mode", self.CMF_multi_cursor
                    )
                    change, self.CMF_highlight_changes = self.menu.checkbox(
                        "Highlight Changes", self.CMF_highlight_changes
                    )

            if self.menu.label_submenu("View"):
                with self.menu.begin_submenu():
                    change, self.CMF_show_line_numbers = self.menu.checkbox(
                        "Show Line Numbers", self.CMF_show_line_numbers
                    )
                    change, self.CMF_show_minimap = self.menu.checkbox(
                        "Show Minimap", self.CMF_show_minimap
                    )
                    change, self.CMF_word_wrap = self.menu.checkbox(
                        "Enable Word Wrap", self.CMF_word_wrap
                    )
                    change, self.CMF_code_folding = self.menu.checkbox(
                        "Enable Code Folding", self.CMF_code_folding
                    )
                    self.menu.separator()
                    if self.menu.label_submenu("Themes"):
                        with self.menu.begin_submenu():
                            change1, self.CMF_theme_dark = self.menu.checkbox(
                                "Dark Mode", self.CMF_theme_dark
                            )
                            change2, self.CMF_theme_light = self.menu.checkbox(
                                "Light Mode", self.CMF_theme_light
                            )
                            change3, self.CMF_theme_solarized = self.menu.checkbox(
                                "Solarized", self.CMF_theme_solarized
                            )
                            change4, self.CMF_theme_custom = self.menu.checkbox(
                                "Use Custom Theme", self.CMF_theme_custom
                            )
                            if change1:
                                self.CMF_theme_light = self.CMF_theme_solarized = (
                                    self.CMF_theme_custom
                                ) = False
                            if change2:
                                self.CMF_theme_dark = self.CMF_theme_solarized = (
                                    self.CMF_theme_custom
                                ) = False
                            if change3:
                                self.CMF_theme_light = self.CMF_theme_dark = (
                                    self.CMF_theme_custom
                                ) = False
                            if change4:
                                self.CMF_theme_light = self.CMF_theme_solarized = (
                                    self.CMF_theme_dark
                                ) = False

            if self.menu.label_submenu("Run"):
                with self.menu.begin_submenu():
                    self.menu.label("Run Current File")
                    self.menu.label("Build All")
                    self.menu.label("Run Tests")
                    if self.menu.label_submenu("Environment"):
                        with self.menu.begin_submenu():
                            self.menu.label("Python 3.10")
                            self.menu.label("Node v18")
                            self.menu.label("Docker")
                            if self.menu.label_submenu("Docker Images"):
                                with self.menu.begin_submenu():
                                    self.menu.label("Alpine")
                                    self.menu.label("Ubuntu")
                                    self.menu.label("Custom Image")
                    self.menu.separator()
                    change, self.CMF_auto_compile = self.menu.checkbox(
                        "Auto Compile", self.CMF_auto_compile
                    )
                    change, self.CMF_show_console = self.menu.checkbox(
                        "Show Console Output", self.CMF_show_console
                    )
                    change, self.CMF_live_reload = self.menu.checkbox(
                        "Live Reload", self.CMF_live_reload
                    )

            if self.menu.label_submenu("Tools"):
                with self.menu.begin_submenu():
                    self.menu.label("Terminal")
                    self.menu.label("Database Explorer")
                    self.menu.label("API Tester")
                    self.menu.label("Profiler")
                    self.menu.label("Debugger")
                    change, self.CMF_devtools = self.menu.checkbox(
                        "Enable DevTools", self.CMF_devtools
                    )
                    change, self.CMF_experimental = self.menu.checkbox(
                        "Show Experimental Tools", self.CMF_experimental
                    )
                    change, self.CMF_logs_enabled = self.menu.checkbox(
                        "Save Logs", self.CMF_logs_enabled
                    )
                    change, self.CMF_verbose_mode = self.menu.checkbox(
                        "Verbose Mode", self.CMF_verbose_mode
                    )

            if self.menu.label_submenu("Extensions"):
                with self.menu.begin_submenu():
                    self.menu.label("Extension Manager")
                    self.menu.label("Marketplace")
                    self.menu.label("Install from File")
                    self.menu.separator()
                    change, self.CMF_extensions_autoupdate = self.menu.checkbox(
                        "Auto-Update Extensions", self.CMF_extensions_autoupdate
                    )
                    change, self.CMF_load_external_plugins = self.menu.checkbox(
                        "Allow External Plugins", self.CMF_load_external_plugins
                    )

            self.menu.separator()
            if self.menu.label_submenu("Account"):
                with self.menu.begin_submenu():
                    self.menu.label("Login")
                    self.menu.label("Logout")
                    change, self.CMF_stay_logged_in = self.menu.checkbox(
                        "Stay Logged In", self.CMF_stay_logged_in
                    )
                    change, self.CMF_cloud_sync = self.menu.checkbox(
                        "Enable Cloud Sync", self.CMF_cloud_sync
                    )

            self.menu.separator()
            change, self.CMF_check_updates = self.menu.checkbox(
                "Check for Updates", self.CMF_check_updates
            )
            change, self.CMF_telemetry = self.menu.checkbox(
                "Send Telemetry Data", self.CMF_telemetry
            )
            change, self.CMF_error_reports = self.menu.checkbox(
                "Send Error Reports", self.CMF_error_reports
            )

            self.menu.separator()
            self.menu.icon_label("Settings", "settings")
            self.menu.icon_label("Help", "help")
            with (
                mili.dict_push(
                    self.menu.style,
                    google_icon_color="red",
                    item_bg_color=(50, 15, 15),
                    hover_color=(100, 30, 30),
                ),
                self.mili.push_styles(text={"color": "red"}),
            ):
                if self.menu.icon_label("Exit", "close"):
                    self.menu.close()


def main():
    MILIUIDemo().run()


if __name__ == "__main__":
    main()



import typing
import pygame
from mili import data as _data

if typing.TYPE_CHECKING:
    from mili import _core
    from mili import utility

__all__ = (
    "SmartNumber",
    "SmartNumberOrPercentage",
    "PointsLike",
    "AnimValueLike",
    "EasingLike",
    "ElementStyleLike",
    "RectStyleLike",
    "CircleStyleLike",
    "LineStyleLike",
    "PolygonStyleLike",
    "TextStyleLike",
    "ImageStyleLike",
    "AnyStyleLike",
    "ScrollbarStyleLike",
    "SliderStyleLike",
    "DropMenuStyleLike",
    "EntryLineStyleLike",
    "MarkDownStyleLike",
    "ComponentProtocol",
    "IconLike",
    "UIAppStyleLike",
    "ContextMenuStyleLike",
    "TransparentRectStyleLike",
    "BoundingAlignLike",
)

type SmartNumber = int | float | str
type SmartNumberOrPercentage = int | float | str
type BoundingAlignLike = typing.Literal[
    "center",
    "topleft",
    "bottomleft",
    "topright",
    "bottomright",
    "midleft",
    "left",
    "midright",
    "right",
    "midtop",
    "top",
    "midbottom",
    "bottom",
]
type PointsLike = typing.Sequence[typing.Sequence[SmartNumberOrPercentage]]
type AnimValueLike = float | pygame.Color | typing.Sequence[float]
type EasingLike = typing.Callable[[float], float]
type IconLike = (
    pygame.Surface
    | typing.Callable[[], pygame.Surface]
    | typing.Callable[[typing.Optional[pygame.typing.ColorLike]], pygame.Surface]
    | str
)


class _ConditionalColorStyleLike(typing.TypedDict):
    default: pygame.typing.ColorLike | None
    hover: pygame.typing.ColorLike | None
    press: pygame.typing.ColorLike | None


class _MarkDownCodeCopyStyleLike(typing.TypedDict):
    size: float
    icon_color: _ConditionalColorStyleLike
    icon_pad: SmartNumberOrPercentage
    icon: IconLike


class _MarkDownCodeScrollbarStyleLike(typing.TypedDict):
    border_radius: float
    height: float
    padx: float
    pady: float
    bar_color: pygame.typing.ColorLike | None
    handle_color: _ConditionalColorStyleLike


class MarkDownStyleLike(typing.TypedDict):
    quote_width: float
    quote_border_radius: float
    indent_width: float
    separator_height: float
    line_break_size: float
    bullet_diameter: float
    table_outline_size: float
    table_border_radius: float
    table_alignx: typing.Literal["left", "center", "right"]
    table_aligny: typing.Literal["top", "center", "bottom"]
    line_break_color: pygame.typing.ColorLike
    bullet_color: pygame.typing.ColorLike
    table_bg_color: pygame.typing.ColorLike | None
    quote_color: pygame.typing.ColorLike
    code_bg_color: pygame.typing.ColorLike
    code_outline_color: pygame.typing.ColorLike
    code_border_radius: SmartNumberOrPercentage
    code_copy_style: _MarkDownCodeCopyStyleLike
    code_scrollbar_style: _MarkDownCodeScrollbarStyleLike
    title1_size: SmartNumber
    title2_size: SmartNumber
    title3_size: SmartNumber
    title4_size: SmartNumber
    title5_size: SmartNumber
    title6_size: SmartNumber
    allow_image_link: bool
    change_cursor: bool
    load_images_async: bool
    link_handler: typing.Callable[[str], None]|None
    parse_async: bool


class _UIAppButtonStyleLike(typing.TypedDict):
    bg_color: _ConditionalColorStyleLike
    icon: IconLike
    alpha: _ConditionalColorStyleLike


class UIAppStyleLike(typing.TypedDict):
    start_style: "ElementStyleLike"
    target_framerate: float
    clear_color: pygame.typing.ColorLike
    window_outline: float
    window_icon: pygame.Surface | None
    outline_color: pygame.typing.ColorLike
    fullscreen_key: int | None
    titlebar_show_title: bool
    title_style: "TextStyleLike"
    titlebar_color: pygame.typing.ColorLike
    minimize_button: _UIAppButtonStyleLike
    maximize_button: _UIAppButtonStyleLike
    close_button: _UIAppButtonStyleLike
    use_appdata_folder: bool


class ScrollbarStyleLike(typing.TypedDict):
    axis: typing.Literal["x", "y"]
    short_size: float
    border_dist: float
    padding: float
    size_reduce: float
    handle_update_id: str | None


class SliderStyleLike(typing.TypedDict):
    lock_x: bool
    lock_y: bool
    handle_size: typing.Sequence[float]
    strict_borders: bool
    area_update_id: str | None
    handle_update_id: str | None
    drag_area: bool


class DropMenuStyleLike(typing.TypedDict):
    direction: typing.Literal["up", "down"]
    anchor: typing.Literal["first", "center", "last"]
    padding: float
    selected_update_id: str | None
    option_update_id: str | None
    menu_update_id: str | None


class DropMenuUIStyleLike(typing.TypedDict):
    border_radius: SmartNumberOrPercentage
    option_border_radius: SmartNumberOrPercentage
    bg_color: pygame.typing.ColorLike
    outline_color: pygame.typing.ColorLike
    selected_option_color: _ConditionalColorStyleLike
    option_color: _ConditionalColorStyleLike
    menu_style: "ElementStyleLike"
    text_style: "TextStyleLike"
    option_text_style: "TextStyleLike"
    icon_arrow_down: IconLike
    icon_arrow_up: IconLike


class EntryLineStyleLike(typing.TypedDict):
    placeholder: str
    placeholder_color: pygame.typing.ColorLike
    text_filly: str
    text_style: "TextStyleLike"
    text_anchor: typing.Literal["left", "right", "center"]
    bg_rect_style: "RectStyleLike|None"
    outline_rect_style: "RectStyleLike|None"
    selection_style: "RectStyleLike|None"
    selection_color: pygame.typing.ColorLike | None
    error_color: pygame.typing.ColorLike
    cursor_color: pygame.typing.ColorLike
    cursor_width: int
    blink_interval: int
    double_click_interval: int
    input_validator: typing.Callable[[str], str | None] | None
    text_validator: typing.Callable[[str], tuple[str, bool]] | None
    validator_lowercase: bool
    validator_uppercase: bool
    validator_windows_path: bool
    target_number: bool
    number_min: float | None
    number_max: float | None
    number_integer: bool
    enable_history: bool
    history_limit: int | None
    keymod: int
    copy_key: int | None
    paste_key: int | None
    cut_key: int | None
    select_all_key: int | None
    delete_all_key: int | None
    undo_key: int | None
    redo_key: int | None
    scroll: "utility.Scroll|None"
    characters_limit: int | None


class _TextBoxScrollbarStyleLike(typing.TypedDict):
    short_size: float
    border_dist: float
    padding: float


class _ContextMenuDefaultIconsStyleLike(typing.TypedDict):
    arrow_left: IconLike
    arrow_right: IconLike
    checkbox_empty: IconLike
    checkbox_full: IconLike


class ContextMenuStyleLike(typing.TypedDict):
    bg_color: pygame.typing.ColorLike
    outline_color: pygame.typing.ColorLike
    hover_color: pygame.typing.ColorLike
    google_icon_color: pygame.typing.ColorLike | typing.Literal["default"] | None
    item_bg_color: pygame.typing.ColorLike | None
    shadow_color: pygame.typing.ColorLike | None
    text_style: "TextStyleLike"
    icon_style: "ImageStyleLike"
    menu_style: "ElementStyleLike"
    row_style: "ElementStyleLike"
    separator_style: "LineStyleLike"
    label_height: float | None
    icon_height: float
    items_update_id: str | None
    menu_border_radius: SmartNumberOrPercentage
    hover_border_radius: SmartNumberOrPercentage
    menu_width: float
    menu_dist: float
    default_icons: _ContextMenuDefaultIconsStyleLike
    submenu_on_hover: bool
    clamp_size: pygame.typing.Point | None


class _ElementStyleLike(typing.TypedDict):
    resizex: bool
    resizey: bool
    fillx: bool | SmartNumberOrPercentage
    filly: bool | SmartNumberOrPercentage
    size_clamp: (
        dict[typing.Literal["min", "max"], tuple[float | None, float | None]] | None
    )
    blocking: bool
    ignore_grid: bool
    align: typing.Literal["first", "center", "last"]
    z: int
    parent_id: int
    offset: typing.Sequence[float]
    clip_draw: bool
    update_id: str
    parent_flag: int
    cache_rect_size: bool
    cache: _data.ParentCache | None

    axis: typing.Literal["x", "y"]
    spacing: SmartNumberOrPercentage
    pad: SmartNumberOrPercentage
    padx: SmartNumberOrPercentage
    pady: SmartNumberOrPercentage
    anchor: typing.Literal["first", "center", "last", "max_spacing"]
    default_align: typing.Literal["first", "center", "last"]
    layout: typing.Literal["stack", "grid", "table"]
    grid_align: typing.Literal[
        "first", "center", "last", "max_spacing", "first_center", "last_center"
    ]
    grid_spacex: SmartNumberOrPercentage
    grid_spacey: SmartNumberOrPercentage


type ElementStyleLike = _ElementStyleLike | dict[str, typing.Any]


class _ComponentStyleLike(typing.TypedDict):
    draw_above: bool
    element_id: int
    pad: SmartNumberOrPercentage
    padx: SmartNumberOrPercentage
    pady: SmartNumberOrPercentage


class _RectStyleLike(_ComponentStyleLike):
    outline: SmartNumberOrPercentage
    border_radius: SmartNumberOrPercentage | typing.Iterable[SmartNumberOrPercentage]
    color: pygame.typing.ColorLike|None
    aspect_ratio: float | None
    align: BoundingAlignLike
    dash_size: SmartNumberOrPercentage | typing.Iterable[SmartNumberOrPercentage] | None
    dash_offset: SmartNumberOrPercentage
    ready_rect: pygame.typing.RectLike | None


type RectStyleLike = _RectStyleLike | dict[str, typing.Any]


class _TransparentRectStyleLike(_ComponentStyleLike):
    border_radius: SmartNumberOrPercentage|pygame.typing.SequenceLike[SmartNumberOrPercentage]
    color: pygame.typing.ColorLike
    alpha: int


type TransparentRectStyleLike = _TransparentRectStyleLike | dict[str, typing.Any]


class _CircleStyleLike(_ComponentStyleLike):
    outline: SmartNumberOrPercentage
    color: pygame.typing.ColorLike|None
    antialias: bool
    aspect_ratio: float | None
    align: BoundingAlignLike
    corners: typing.Iterable[bool]
    dash_size: float | typing.Iterable[float] | None
    dash_anchor: float | typing.Iterable[float]


type CircleStyleLike = _CircleStyleLike | dict[str, typing.Any]


class _LineStyleLike(_ComponentStyleLike):
    size: SmartNumberOrPercentage
    color: pygame.typing.ColorLike|None
    antialias: bool
    dash_size: SmartNumberOrPercentage | typing.Iterable[SmartNumberOrPercentage]
    dash_offset: SmartNumberOrPercentage


type LineStyleLike = _LineStyleLike | dict[str, typing.Any]


class _PolygonStyleLike(_ComponentStyleLike):
    outline: SmartNumberOrPercentage
    color: pygame.typing.ColorLike|None


type PolygonStyleLike = _PolygonStyleLike | dict[str, typing.Any]


class _TextRichMarkdownStyleLike(typing.TypedDict):
    spoiler_color: pygame.typing.ColorLike
    spoiler_text_color: pygame.typing.ColorLike
    code_bg_color: pygame.typing.ColorLike
    code_text_color: pygame.typing.ColorLike
    highlight_color: pygame.typing.ColorLike


class _TextStyleLike(_ComponentStyleLike):
    cache: _data.TextCache | None
    name: str | None
    size: SmartNumber
    sysfont: bool
    align: BoundingAlignLike
    font_align: int
    font_direction: int
    bold: bool
    italic: bool
    underline: bool
    strikethrough: bool
    antialias: bool
    color: pygame.typing.ColorLike
    bg_color: pygame.typing.ColorLike
    outline_color: pygame.typing.ColorLike
    growx: bool
    growy: bool
    wraplen: SmartNumberOrPercentage
    slow_grow: bool
    rich: bool
    rich_aligny: typing.Literal["top", "center", "bottom"]
    rich_linespace: int
    rich_actions: dict[str, typing.Callable[[typing.Any], None]]
    rich_link_color: pygame.typing.ColorLike
    rich_markdown: bool
    rich_markdown_style: _TextRichMarkdownStyleLike


type TextStyleLike = _TextStyleLike | dict[str, typing.Any]


class _ImageStyleLike(_ComponentStyleLike):
    cache: _data.ImageCache | None
    layer_cache: _data.ImageLayerCache | None
    fill: bool
    stretchx: bool
    stretchy: bool
    fill_color: pygame.typing.ColorLike
    smoothscale: bool
    ready_border_radius: SmartNumberOrPercentage
    border_radius: (
        SmartNumberOrPercentage | pygame.typing.SequenceLike[SmartNumberOrPercentage]
    )
    alpha: int
    ninepatch_size: SmartNumberOrPercentage
    ready: bool
    transforms: (
        pygame.typing.SequenceLike[typing.Callable | pygame.typing.SequenceLike] | None
    )
    filters: (
        pygame.typing.SequenceLike[typing.Callable | pygame.typing.SequenceLike] | None
    )


type ImageStyleLike = _ImageStyleLike | dict[str, typing.Any]


type AnyStyleLike = (
    ElementStyleLike
    | RectStyleLike
    | CircleStyleLike
    | LineStyleLike
    | PolygonStyleLike
    | TextStyleLike
    | ImageStyleLike
)


class ComponentProtocol(typing.Protocol):
    def draw(
        self,
        ctx: "_core._ctx",
        data: typing.Any,
        style: dict[str, typing.Any],
        element: dict[str, typing.Any],
        absolute_rect: pygame.Rect,
    ): ...

    def added(
        self,
        ctx: "_core._ctx",
        data: typing.Any,
        style: dict[str, typing.Any],
        element: dict[str, typing.Any],
    ): ...


# deprecated. will be removed in 1.0.8
type NumberOrPercentage = SmartNumberOrPercentage



import pygame
import typing
from mili import _core
from mili import data as _data


class InteractionSound:
    def __init__(
        self,
        hover: pygame.mixer.Sound | None = None,
        unhover: pygame.mixer.Sound | None = None,
        press: pygame.mixer.Sound | None = None,
        release: pygame.mixer.Sound | None = None,
        button: int = pygame.BUTTON_LEFT,
        update_id: str | None = None,
    ):
        self.hover_sound = hover
        self.unhover_sound = unhover
        self.press_sound = press
        self.release_sound = release
        self.button = button
        _core._globalctx._register_update_id(update_id, self.play)

    def play(
        self,
        interaction: _data.Interaction,
        channel: pygame.Channel | None = None,
        maxtime: int = 0,
        fade_ms: int = 0,
    ) -> _data.Interaction:
        if interaction.just_hovered and self.hover_sound is not None:
            if channel is not None:
                channel.play(self.hover_sound, 0, maxtime, fade_ms)
            else:
                self.hover_sound.play(0, maxtime, fade_ms)
        if interaction.just_unhovered and self.unhover_sound is not None:
            if channel is not None:
                channel.play(self.unhover_sound, 0, maxtime, fade_ms)
            else:
                self.unhover_sound.play(0, maxtime, fade_ms)
        if (
            interaction.just_pressed_button == self.button
            and self.press_sound is not None
        ):
            if channel is not None:
                channel.play(self.press_sound, 0, maxtime, fade_ms)
            else:
                self.press_sound.play(0, maxtime, fade_ms)
        if (
            interaction.just_released_button == self.button
            and self.release_sound is not None
        ):
            if channel is not None:
                channel.play(self.release_sound, 0, maxtime, fade_ms)
            else:
                self.release_sound.play(0, maxtime, fade_ms)
        return interaction


class InteractionCursor:
    _cursor_stolen: int | pygame.Cursor | None = None
    idle_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_ARROW
    hover_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_HAND
    press_cursor: int | pygame.Cursor | None | dict[int, int | pygame.Cursor | None] = (
        pygame.SYSTEM_CURSOR_HAND
    )
    disabled_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_NO
    _active = True

    @classmethod
    def setup(
        cls,
        idle_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_ARROW,
        hover_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_HAND,
        press_cursor: int
        | pygame.Cursor
        | None
        | dict[int, int | pygame.Cursor | None] = pygame.SYSTEM_CURSOR_HAND,
        disabled_cursor: int | pygame.Cursor | None = pygame.SYSTEM_CURSOR_NO,
        update_id: str | None = "cursor",
    ):
        cls._cursor_stolen = None
        cls.idle_cursor = idle_cursor
        cls.hover_cursor = hover_cursor
        cls.press_cursor = press_cursor
        cls.disabled_cursor = disabled_cursor
        cls._active = True
        _core._globalctx._register_update_id(update_id, cls.update)

    @classmethod
    def apply(
        cls,
    ) -> int | pygame.Cursor | None:
        if not cls._active:
            cls._cursor_stolen = None
            return None
        if cls._cursor_stolen is None and cls.idle_cursor is not None:
            pygame.mouse.set_cursor(cls.idle_cursor)
            cls._cursor_stolen = None
            return cls.idle_cursor
        elif cls._cursor_stolen is not None:
            pygame.mouse.set_cursor(cls._cursor_stolen)
            cursor_ret = cls._cursor_stolen
            cls._cursor_stolen = None
            return cursor_ret
        cls._cursor_stolen = None
        return None

    @classmethod
    def set_status(cls, status: typing.Literal["hover", "press", "disabled", "idle"]):
        match status:
            case "hover":
                cls._cursor_stolen = cls.hover_cursor
            case "press":
                if isinstance(cls.press_cursor, dict):
                    cls._cursor_stolen = cls.press_cursor[
                        list(cls.press_cursor.keys())[0]
                    ]
                else:
                    cls._cursor_stolen = cls.press_cursor
            case "disabled":
                cls._cursor_stolen = cls.disabled_cursor
            case "idle":
                cls._cursor_stolen = None

    @classmethod
    def set_cursor(cls, cursor):
        cls._cursor_stolen = cursor

    @classmethod
    def update(
        cls,
        interaction: _data.Interaction,
        disabled: bool = False,
        **cursor_overrides: int
        | pygame.Cursor
        | None
        | dict[int, int | pygame.Cursor | None],
    ) -> _data.Interaction:
        hovered = interaction.hovered
        pressed = interaction.press_button != -1
        if disabled and (hovered or pressed):
            cursor = cursor_overrides.get("disabled_cursor", cls.disabled_cursor)
            if not isinstance(cursor, dict):
                cls._cursor_stolen = cursor
            else:
                cls._cursor_stolen = cls.disabled_cursor
            if cls._cursor_stolen is not None:
                return interaction
        if pressed:
            press_cursor = cursor_overrides.get("press_cursor", cls.press_cursor)
            if isinstance(press_cursor, dict):
                cls._cursor_stolen = press_cursor.get(interaction.press_button, None)
                if cls._cursor_stolen is not None:
                    return interaction
            elif interaction.press_button == pygame.BUTTON_LEFT:
                cls._cursor_stolen = press_cursor
                if cls._cursor_stolen is not None:
                    return interaction
        if hovered:
            cursor = cursor_overrides.get("hover_cursor", cls.hover_cursor)
            if not isinstance(cursor, dict):
                cls._cursor_stolen = cursor
            else:
                cls._cursor_stolen = cls.hover_cursor
            return interaction
        return interaction
